from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from .exceptions import InvalidSectionNumberForCourseError
from .enums import SemesterType
from .section import Section

# stuff that we need just for type checking, not for actual functionality
if TYPE_CHECKING:
    from .block import Block
    from .teacher import Teacher
    from .stream import Stream
    from .lab import Lab

DEFAULT_HOURS: float = 3.0

OptionalId = Optional[int]


class Course:
    """Describes a distinct course."""

    # -------------------------------------------------------------------
    # new
    # --------------------------------------------------------------------
    def __init__(self, number: str = "",
                 name: str = "New Course",
                 semester: SemesterType = SemesterType.any,
                 hours_per_week: float = 3.0,
                 needs_allocation: bool = True):
        """Creates and returns a course object.
        :param number: Course number
        :param name: Course name
        :param semester: Which semester is it taught in
        :param needs_allocation: This requires someone to teach it
        :course_id: if not specified, will create one as required
        """

        self._number: str = number
        self.name: str = name
        self.needs_allocation: bool = needs_allocation
        self.hours_per_week = float(hours_per_week)
        self._sections: set[Section] = set()
        self.semester: SemesterType = semester

    # =================================================================
    # unique identifier
    # =================================================================
    @property
    def number(self) -> str:
        """Returns the unique ID for this Course object."""
        return self._number

    # =================================================================
    # title
    # =================================================================
    @property
    def title(self) -> str:
        """Returns text string that describes this Block."""
        return f"{self.number}: {self.name}"

    # =================================================================
    # sections
    # =================================================================
    def add_section(self, number: str = "",  name: str = "",
                    section_id: int = None) -> Section:
        """Assign a Section to this Course."""

        if number in (s.number for s in self.sections()):
            raise InvalidSectionNumberForCourseError(
                f"<{number}>: section number is not unique for this Course.")

        if number == "":
            max_number = 0
            for sec_number in (s.number for s in self.sections()):
                try:
                    max_number = max(int(sec_number), max_number)
                except ValueError:
                    pass
            number = str(max_number + 1)

        section = Section(self, number, name, section_id)
        self._sections.add(section)

        return section

    def get_section_by_number(self, number: str) -> Optional[Section]:
        """Gets the Section from this Course that has the passed section number, if it exists.
        Otherwise, returns None. """
        found = [section for section in self.sections() if section.number == number]
        return found[0] if found else None

    def get_section_by_id(self, sect_id: int) -> Optional[Section]:
        """Gets the Section from this Course that matches the passed section ID, if it exists.
        Returns the Section if found, or None otherwise."""
        found = [section for section in self.sections() if section.id == sect_id]
        return found[0] if found else None

    def get_sections_by_name(self, name: str) -> tuple[Section, ...]:
        """Gets the Section from this Course that has the passed section name, if it exists.
        Returns a list containing the Section if found, or an empty list otherwise."""
        found = (section for section in self.sections() if section.name == name)
        return tuple(found)

    def remove_section(self, section: Section):
        """Removes the passed Section from this Course, if it exists."""
        self._sections.discard(section)

    def sections(self) -> tuple[Section, ...]:
        """Returns a list of all the Sections assigned to this Course."""
        return tuple(sorted(self._sections))

    def number_of_sections(self) -> int:
        """Returns the number of Sections assigned to this Course."""
        return len(self.sections())

    def get_sections_for_teacher(self, teacher: Teacher) -> tuple[Section, ...]:
        """Returns a list of the Sections assigned to this course, with this Teacher."""
        sections = []

        for section in self.sections():
            if teacher in section.teachers():
                sections.append(section)

        return tuple(sorted(set(sections)))

    def get_sections_for_allocated_teacher(self, teacher: Teacher) -> tuple[Section, ...]:
        """Returns a list of the Sections assigned to this course, with this Teacher."""
        sections = []

        for section in self.sections():
            if section.has_allocated_teacher(teacher):
                sections.append(section)

        return tuple(sorted(set(sections)))

    def max_section_number(self) -> int:
        """Returns the maximum Section number from the Sections assigned to this Course."""
        sections = sorted(self.sections(), key=lambda s: s.number)
        return sections[-1].number if len(sections) > 0 else 0

    def remove_all_sections(self):
        """remove all sections in this course"""
        for section in self.sections():
            self.remove_section(section)

    # =================================================================
    # blocks
    # =================================================================
    def blocks(self) -> tuple[Block, ...]:
        """Returns a tuple of the Blocks assigned to this Course."""
        blocks = []
        for section in self.sections():
            blocks.extend(section.blocks())
        return tuple(sorted(set(blocks)))

    # =================================================================
    # teachers
    # =================================================================
    def teachers(self) -> tuple[Teacher, ...]:
        """Returns a list of the Teachers assigned to all Sections of this Course."""

        teachers: list[Teacher] = list()
        for section in self.sections():
            teachers.extend(section.teachers())
        return tuple(sorted(set(teachers), key=lambda x: (x.lastname, x.firstname)))

    def has_teacher(self, teacher: Teacher) -> bool:
        """Returns true if the passed Teacher is assigned to this Course."""
        return teacher in self.teachers()

    def has_allocated_teacher(self, teacher: Teacher) -> bool:
        """Returns true if the passed Teacher is assigned to this Course."""
        if teacher in self.teachers():
            return True
        for section in self.sections():
            if section.has_allocated_teacher(teacher):
                return True
        return False

    def add_teacher(self, teacher: Teacher) -> Teacher:
        """Assigns a Teacher to all Sections of this Course."""

        for section in self.sections():
            section.add_teacher(teacher)
        return teacher

    def remove_teacher(self, teacher: Teacher):
        """Removes the passed Teacher from all Sections/Blocks in this Course."""
        for section in self.sections():
            section.remove_teacher(teacher)

    def remove_all_teachers(self):
        """Removes all Teachers from all Sections/Blocks in this Course."""
        for teacher in self.teachers():
            self.remove_teacher(teacher)


    # =================================================================
    # stream
    # =================================================================
    def streams(self) -> tuple[Stream, ...]:
        """Returns a list of Streams assigned to all Sections of this Course."""

        streams: list[Stream] = list()
        for section in self.sections():
            streams.extend(section.streams())
        return tuple(sorted(set(streams)))

    def has_stream(self, stream: Stream) -> bool:
        """Returns true if this Course has the specified Stream."""
        return stream in self.streams()

    def add_stream(self, stream: Stream) -> Stream:
        """Assigns a Stream to all Sections of this Course."""
        if stream:
            for section in self.sections():
                section.add_stream(stream)

        return stream

    def remove_stream(self, stream: Stream):
        """Removes the specified Stream from this Course."""
        for section in self.sections():
            section.remove_stream(stream)

    def remove_all_streams(self):
        """Removes all Streams from all Sections of this Course."""
        for stream in self.streams():
            self.remove_stream(stream)



    # =================================================================
    # labs
    # =================================================================
    def labs(self) -> tuple[Lab, ...]:
        """Returns a list of Labs assigned to all Sections of this Course."""

        lab_list: list[Lab] = []

        for section in self.sections():
            for block in section.blocks():
                lab_list.extend(block.labs())
        return tuple(sorted(set(lab_list)))

    def add_lab(self, lab: Lab) -> Lab:
        """Assigns a Lab to all Sections of this Course."""

        for section in self.sections():
            section.add_lab(lab)
        return lab

    def has_lab(self, lab: Lab):
        """Does this course have this lab in any of its sections/blocks?"""
        return lab in self.labs()

    def remove_lab(self, lab: Lab):
        for section in self.sections():
            section.remove_lab(lab)

    def remove_all_labs(self):
        for lab in self.labs():
            self.remove_lab(lab)

    # =================================================================
    # Get unused section number
    # =================================================================
    def get_new_number(self, number: int = 1) -> int:
        """Returns the first unused Section Number."""
        while self.get_section_by_number(str(number)):
            number += 1
        return number

    # ------------------------------------------------------------------------
    # for sorting
    # ------------------------------------------------------------------------
    def __lt__(self, other):
        return self.number < other.number

    def __hash__(self):
        return hash(self._number)

    def __eq__(self, other):
        return self.number == other.number

    # =================================================================
    # string representation of the object
    # =================================================================
    def __str__(self) -> str:
        """Returns a text string that describes the Course, its Sections, its Blocks,
        its Teachers, and its Labs. """
        text = ''

        # Header
        text += "\n\n"
        text += '=' * 50 + "\n"
        text += f"{self._number} {self.name}\n"
        text += '=' * 50 + "\n"

        # Sections
        for s in sorted(self.sections(), key=lambda x: x.number):
            text += f"\n{s}\n"
            text += "-" * 50 + "\n"

            # Blocks
            for b in sorted(s.blocks()):
                text += f"{b}\n"
                text += "\tlab_ids: " + ", ".join([str(lab) for lab in b.labs()]) + "\n"
                text += "\tteacher_ids: "
                text += ", ".join([str(t) for t in b.teachers()])
                text += "\n"

        return text

    def __repr__(self) -> str:
        return f"{self.title}  ({self.hours_per_week} hrs/wk)"




from typing import Optional

from . import FindImages
from tkinter import Frame, PhotoImage, Button, Label
import os
from functools import partial
from .idlelib_tooltip import Hovertip
from .InitGuiFontsAndColours import TkColours
import os

BUTTON_SIZE=24
BUTTON_BORDERWIDTH = 1


# TODO: Standardize the method for cget/configure between TableEntry,
#       dynamic entry, and ToolBar.  (Best? is DynamicTree)

class ToolBar(Frame):
    def __init__(self, parent, colours: Optional[TkColours] = None, **kwargs):
        super().__init__(master=parent, **kwargs)

        # define top level for this frame
        mw = self.master.winfo_toplevel()

        # set up default colours
        self.mw = mw
        if colours is not None:
            self.colours = colours
        else:
            self.colours = TkColours(mw)

        # default configuration
        self.config_options = ['button_hover_highlight', 'button_highlight', 'button_fg', 'button_bg']
        self.configure(bg=colours.WorkspaceColour)
        self.configure(button_hover_highlight=colours.ButtonHoverHighlight)
        self.configure(button_highlight=colours.WorkspaceColour)
        self.configure(button_bg=colours.ButtonBackground)
        self.configure(button_fg=colours.ButtonForeground)

        # user configuration
        for k,v in kwargs.items():
            self.configure(**{k:v})

        self.images = list()

        # some defaults
        image_dir = FindImages.get_image_dir()

        self.defaults = {
            'image': os.path.join(image_dir, 'default.gif')
        }

        # buttons
        self.buttons: dict[str, Button] = {}

    def configure(self, **kwargs):
        # https://stackoverflow.com/a/70568930
        # workaround to tkinter seemingly not having ConfigSpecs
        for key in self.config_options:
            if key in kwargs:
                setattr(self, key, kwargs.pop(key))
        if kwargs:
            super().configure(**kwargs)

    def config(self, **kwargs):
        self.configure(**kwargs)

    def cget(self, key):
        # if key in self.config_options:
        #     print (f"{key=}")
        return getattr(self, key, None) if key in self.config_options \
            else super().cget(key)

    def add(self, **details):
        """Adds a button

        ---
        Parameters:
            - name     : str
            - image    : str (Path)
            - command  : method
            - hint     : str
            - shortcut : str
            - disabled : bool
        """

        btn_height = 20
        btn_width = 20

        # define inputs and defaults
        image_file = details.get('image', '')
        if not os.path.exists(image_file) or not os.path.isfile(image_file):
            image_file = self.defaults['image']

        name = details.get('name', os.path.basename(image_file))
        callback: callable = details.get('command', lambda *_, **__: None)
        hint = details.get('hint', '')
        shortcut_key = details.get('shortcut', [])
        disabled = details.get('disabled', False)

        # define image
        image = PhotoImage(file=image_file)

        # if the image is larger than the button, roughly scale it down, so it's not auto-cropped
        # NOTE: have to save the image in object instance so that the ref does not get removed
        if btn_width < image.width() or btn_height < image.height():
            image = image.subsample(round(image.width() / btn_width),
                                    round(image.height() / btn_height))
        self.images.append(image)

        # add button
        b = Button(
            master=self,
            text=name,
            image=image,
            relief='ridge',
            bg=self.cget("button_bg"),
            activebackground=self.cget("button_bg"),
            highlightbackground=self.cget("button_highlight"),

            fg=self.cget("button_fg"),
            width=BUTTON_SIZE,
            height=BUTTON_SIZE,
            highlightthickness=BUTTON_BORDERWIDTH,
        )
        b.pack(side='left',ipadx=0, ipady=0)
        self.buttons[name] = b

        # weird behaviour, using 'command=' doesn't work, but this does
        b.bind('<Button-1>', callback)

        # add the tooltip
        if hint:
            Hovertip(b, hint, 1000)

        # make the button slightly raised when mouse hovers
        # pass image to stop it from being destroyed when this method ends
        b.bind('<Enter>', partial(button_enter, b, self.cget("button_hover_highlight"), self.mw, image), True)
        b.bind('<Leave>', partial(button_leave, b, self.cget("button_highlight"), self.mw, image), True)

        # add the bindings for key presses
        if shortcut_key:
            for sk in shortcut_key:
                self.mw.bind(f"<{sk}>", partial(lambda *_: b.invoke(), b))

        # disable the button?
        if disabled:
            b.configure(state='disabled')

    def bar(self):
        """Add a divider on the _toolbar"""
        Label(self, text='', bg=self.cget('bg'), relief='raised', borderwidth=1) \
            .pack(side='left', padx=3)

    def enable(self, button):
        """Enable a button"""
        self._state(1, button)

    def disable(self, button):
        """Disable a button"""
        self._state(0, button)

    def _state(self, action, name):
        if name in self.buttons and isinstance(self.buttons[name], Button):
            self.buttons[name].configure(state='normal' if action else 'disabled')


def button_enter(b: Button, highlight_color, mw, *_):
    b.configure(highlightbackground=highlight_color)
    mw.update()


def button_leave(b: Button, highlight_color, mw, *_):
    b.configure(highlightbackground = highlight_color)
    mw.update()



import pytest

from src.scheduling_and_allocation.model import ConflictType, enums, ResourceType, Block, WeekDay, \
                set_block_conflicts, set_lunch_break_conflicts, \
                set_number_of_days_conflict, MAX_HOURS_PER_WEEK, set_availability_hours_conflict


class ParentContainer:
    @property
    def id(self) -> int:
        return 1

    @property
    def title(self) -> str:
        return "whatever"


parent = ParentContainer

ALL_BITS_SET = ~ConflictType(0)


# ============================================================================
# verify that the is_??_conflict(int) works as expected
# ============================================================================
def test_is_time():
    cc = ConflictType.TIME
    assert cc.is_time()
    cc = ConflictType.TIME | ConflictType.TIME_LAB | ConflictType.NONE | ConflictType.LUNCH
    assert cc.is_conflicted()
    assert cc.is_time()
    cc = cc ^ ConflictType.TIME
    assert not cc.is_time()


def test_is_time_lab():
    cc = ConflictType.TIME_LAB
    assert cc.is_time_lab()
    cc = ConflictType.TIME_LAB | ConflictType.TIME | ConflictType.NONE | ConflictType.LUNCH
    assert cc.is_time_lab()
    cc = cc ^ ConflictType.TIME_LAB
    assert not cc.is_time_lab()


def test_is_time_stream():
    cc = ConflictType.TIME_STREAM
    assert cc.is_time_stream()
    cc = ConflictType.TIME_STREAM | ConflictType.TIME | ConflictType.NONE | ConflictType.LUNCH
    assert cc.is_time_stream()
    cc = cc ^ ConflictType.TIME_STREAM
    assert not cc.is_time_stream()


def test_is_time_teacher():
    cc = ConflictType.TIME_TEACHER
    assert cc.is_time_teacher()
    cc = ConflictType.TIME_TEACHER | ConflictType.TIME | ConflictType.NONE | ConflictType.LUNCH
    assert cc.is_time_teacher()
    cc = cc ^ ConflictType.TIME_TEACHER
    assert not cc.is_time_teacher()


def test_is_time_lunch():
    cc = ConflictType.LUNCH
    assert cc.is_time_lunch()
    cc = ConflictType.TIME | ConflictType.TIME | ConflictType.NONE | ConflictType.LUNCH
    assert cc.is_time_lunch()
    cc = cc ^ ConflictType.LUNCH
    assert not cc.is_time_lunch()


def test_is_minimum_days():
    cc = ConflictType.MINIMUM_DAYS
    assert cc.is_minimum_days()
    cc = ConflictType.TIME | ConflictType.TIME | ConflictType.NONE | ConflictType.MINIMUM_DAYS
    assert cc.is_minimum_days()
    cc = cc ^ ConflictType.MINIMUM_DAYS
    assert not cc.is_minimum_days()


def test_is_availability():
    cc = ConflictType.AVAILABILITY
    assert cc.is_availability()
    cc = ConflictType.TIME | ConflictType.TIME | ConflictType.NONE | ConflictType.AVAILABILITY
    assert cc.is_availability()
    cc = cc ^ ConflictType.AVAILABILITY
    assert not cc.is_availability()


# ============================================================================
# most severe
# ============================================================================
severity_order = enums.ORDER_OF_SEVERITY

def test_severity_dependent_on_view():
    """if view can have a conflict specific to its view, then it should be the most severe"""
    cc = ALL_BITS_SET

    most_severe: ConflictType = cc.most_severe(ResourceType.lab)
    assert most_severe == ConflictType.TIME_LAB

    most_severe: ConflictType = cc.most_severe(ResourceType.stream)
    assert most_severe == ConflictType.TIME_STREAM

    most_severe: ConflictType = cc.most_severe(ResourceType.teacher)
    assert most_severe == ConflictType.TIME_TEACHER

    cc = ConflictType.NONE
    most_severe: ConflictType = cc.most_severe( ResourceType.teacher)
    assert most_severe == ConflictType.NONE

    most_severe: ConflictType = cc.most_severe( ResourceType.lab)
    assert most_severe == ConflictType.NONE

    most_severe: ConflictType = cc.most_severe( ResourceType.stream)
    assert most_severe == ConflictType.NONE

    cc = ALL_BITS_SET
    most_severe: ConflictType = cc.most_severe( ResourceType.none)
    assert most_severe != ConflictType.TIME_TEACHER and \
           most_severe != ConflictType.TIME_STREAM and \
           most_severe != ConflictType.TIME_LAB


def test_severity_view_independent():
    """Verity that returned conflict resource_type is the most severe"""

    cc = ConflictType.NONE
    for ct in severity_order:
        cc = ct | cc

    for ct in severity_order:
        most_severe = cc.most_severe(ResourceType.none)
        assert most_severe == ct
        cc = cc ^ ct

    cc = ConflictType.NONE
    most_severe = cc.most_severe(ResourceType.none)
    assert most_severe == ConflictType.NONE


# ============================================================================
# time conflicts
# ============================================================================
def test_block_time_conflicts_is_overlap():
    """Detects a block overlap"""
    block1 = Block(parent, WeekDay.Monday, 8.5, 2)
    block2 = Block(parent, WeekDay.Monday,12, 2)
    block3 = Block(parent, WeekDay.Monday, 10, 2)

    set_block_conflicts((block1, block2, block3), ConflictType.TIME_LAB)
    assert block1.conflict.is_time_lab()
    assert block3.conflict.is_time_lab()
    assert block1.conflict.is_time()
    assert block3.conflict.is_time()
    assert not block2.conflict.is_conflicted()


def test_block_time_with_overlap_on_different_day():
    """determines no overlap"""
    block1 = Block(parent, WeekDay.Monday, 8.5, 2)
    block2 = Block(parent, WeekDay.Monday, 12, 2)
    block3 = Block(parent, WeekDay.Tuesday,10, 2)

    set_block_conflicts((block1, block2, block3), ConflictType.TIME_LAB)
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()


# ============================================================================
# lunch breaks (need 1/2 hour from 11am until 2pm)
# ============================================================================
def test_lunch_breaks_single_block():
    """Verifies that there is a lunch if a single block"""

    # single block spans entire lunch break
    block1 = Block(parent, WeekDay.Monday, 10.5, 5)
    set_lunch_break_conflicts((block1,))
    assert block1.conflict.is_time_lunch()

    # single block that still gives lunch break
    block1 = Block(parent, WeekDay.Monday, 11.5, 5)
    set_lunch_break_conflicts((block1,))
    assert not block1.conflict.is_conflicted()


def test_lunch_breaks_two_blocks_with_break_inbetween():
    """two blocks with whole lunch break taken, except for 1/2 in between"""
    block1 = Block(parent, WeekDay.Monday, 10.5, 2)
    block2 = Block(parent, WeekDay.Monday, 13, 5)
    block3 = Block(parent, WeekDay.Tuesday, 12.5, 1.5)
    set_lunch_break_conflicts((block1, block2, block3))
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()


def test_lunch_three_blocks_with_no_breaks():
    """three blocks with no breaks"""
    block1 = Block(parent, WeekDay.Monday, 10.5, 2)
    block2 = Block(parent, WeekDay.Monday, 13, 5)
    block3 = Block(parent, WeekDay.Monday, 12.5, 1.5)
    set_lunch_break_conflicts((block1, block2, block3))
    assert block1.conflict.is_time_lunch()
    assert block2.conflict.is_time_lunch()
    assert block3.conflict.is_time_lunch()


def test_lunch_three_blocks_with_break_at_beginning():
    """three blocks with break at beginning"""
    block1 = Block(parent, WeekDay.Monday, 11.5, 1)
    block2 = Block(parent, WeekDay.Monday, 13, 5)
    block3 = Block(parent, WeekDay.Monday, 12.5, 1.5)
    set_lunch_break_conflicts((block1, block2, block3))
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()


def test_lunch_three_blocks_with_break_at_end():
    """three blocks with break at end"""
    block1 = Block(parent, WeekDay.Monday, 10.5, 2)
    block2 = Block(parent, WeekDay.Monday, 13, 0.5)
    block3 = Block(parent, WeekDay.Monday, 12.5, 1.5)
    set_lunch_break_conflicts((block1, block2, block3))
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()


# ============================================================================
# number_of_days
# ============================================================================
def test_number_of_days_too_little():
    block1 = Block(parent, WeekDay.Monday, 10.5, 2)
    block2 = Block(parent, WeekDay.Tuesday, 13, 0.5)
    block3 = Block(parent, WeekDay.Wednesday, 12.5, 1.5)
    block4 = Block(parent, WeekDay.Monday, 12.5, 1.5)
    set_number_of_days_conflict((block1, block2, block3, block4))
    assert block1.conflict.is_minimum_days()
    assert block2.conflict.is_minimum_days()
    assert block3.conflict.is_minimum_days()
    assert block4.conflict.is_minimum_days()


def test_number_of_days_ok():
    block1 = Block(parent, WeekDay.Monday,10.5, 2)
    block2 = Block(parent, WeekDay.Tuesday, 13.0, 0.5)
    block3 = Block(parent, WeekDay.Wednesday, 12.5, 1.5)
    block4 = Block(parent, WeekDay.Thursday, 12.5, 1.5)
    set_number_of_days_conflict((block1, block2, block3, block4))
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()
    assert not block4.conflict.is_conflicted()


# ============================================================================
# allocated hours
# ============================================================================
def test_too_many_hours_per_week():
    block1 = Block(parent, WeekDay.Monday, 8.5, 2)
    block2 = Block(parent, WeekDay.Monday, 14, 2)
    block3 = Block(parent, WeekDay.Tuesday,8.5, 1.5)
    block4 = Block(parent, WeekDay.Tuesday, 14, 2)
    block5 = Block(parent, WeekDay.Wednesday, 8.5, 1.5)
    block6 = Block(parent, WeekDay.Wednesday, 17.5, 1.5)
    block7 = Block(parent, WeekDay.Thursday, 8.5, 1.5)
    block8 = Block(parent, WeekDay.Thursday, 17.5, 1.5)
    # verify that we got the right conditions before testing starting and stopping late
    total_hours = ((block2.end - block1.start)
                   + (block4.end - block3.start)
                   + (block6.end - block5.start)
                   + (block8.end - block7.start))

    assert total_hours > MAX_HOURS_PER_WEEK
    set_availability_hours_conflict((block1, block2, block3, block4, block5, block6, block7, block8))
    assert block1.conflict.is_availability()
    assert block2.conflict.is_availability()
    assert block3.conflict.is_availability()
    assert block4.conflict.is_availability()
    assert block5.conflict.is_availability()
    assert block6.conflict.is_availability()
    assert block7.conflict.is_availability()
    assert block8.conflict.is_availability()


def test_ok_many_hours_per_week():
    block1 = Block(parent, WeekDay.Monday, 8.5, 2)
    block2 = Block(parent, WeekDay.Monday, 14, 2)
    block3 = Block(parent, WeekDay.Tuesday, 8.5, 1.5)
    block4 = Block(parent, WeekDay.Tuesday, 14, 2)
    block5 = Block(parent, WeekDay.Wednesday, 8.5, 1.5)
    block6 = Block(parent, WeekDay.Wednesday, 17.5, 1.5)
    block7 = Block(parent, WeekDay.Thursday, 8.5, 1.5)
    block8 = Block(parent, WeekDay.Thursday, 10.5, 1.5)

    # verify that we got the right conditions before testing starting and stopping late
    total_hours = ((block2.end - block1.start)
                   + (block4.end - block3.start)
                   + (block6.end - block5.start)
                   + (block8.end - block7.start))

    assert total_hours < MAX_HOURS_PER_WEEK
    set_availability_hours_conflict((block1, block2, block3, block4, block5, block6, block7, block8))
    assert not block1.conflict.is_conflicted()
    assert not block2.conflict.is_conflicted()
    assert not block3.conflict.is_conflicted()
    assert not block4.conflict.is_conflicted()
    assert not block5.conflict.is_conflicted()
    assert not block6.conflict.is_conflicted()
    assert not block7.conflict.is_conflicted()
    assert not block8.conflict.is_conflicted()



"""
# ============================================================================
# MAIN PAGE - Gui entry point to the application(s)
#
# - Forms base class for Scheduler and Allocation applications
#
# EVENT HANDLERS - None
# ============================================================================

"""
from __future__ import annotations

import platform
from functools import partial

import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog as tk_fd
from typing import Optional, TYPE_CHECKING, Callable

from tkinter.messagebox import showerror, showinfo, askyesno

from ..modified_tk import set_default_fonts_and_colours, TkColours, TkFonts
from ..gui_dialogs.change_font_tk import ChangeFont
from ..gui_pages.note_book_frame_tk import NoteBookFrameTk, TabInfoProtocol
from ..gui_generics.menu_and_toolbars import MenuItem, ToolbarItem, generate_menu, make_toolbar
from ..Utilities.Preferences import Preferences
from ..modified_tk.InitGuiFontsAndColours import set_system_colours

if TYPE_CHECKING:
    pass

Operating_system = platform.system().lower()

MAIN_FRAME_HEIGHT = 600
MAIN_FRAME_WIDTH = 800
WELCOME_WIDTH = 800
WELCOME_HEIGHT = 600


class MainPageBaseTk:
    """Generic main page for Scheduler and Allocation Manager apps"""
    bin_dir = None

    # ===================================================================================
    # constructor
    # ===================================================================================
    def __init__(self, title, preferences: Preferences):
        """
        main page
        :param title:
        :param preferences: a preference object (color scheme, last opened file, etc)
        """

        # public properties
        self.dark_mode = preferences.dark_mode()
        self.theme = preferences.theme()
        self.colours: Optional[TkColours] = None
        self.fonts: Optional[TkFonts] = None
        self.dict_of_frames: dict[str, tk.Frame] = dict()
        self.logo: Optional[tk.PhotoImage] = None
        self.mw: Optional[tk.Tk] = None
        self.notebook_tab_changed_handler: Callable[[str,tk.Frame], None] = lambda a,b: None

        # autosave handler
        self.toggle_auto_save: Callable[ [bool], None] = lambda flag: print("Auto set has been set to", flag)

        # private properties
        self._preferences = preferences
        self._notebook_frame: Optional[NoteBookFrameTk] = None
        self._wait = None
        self.notebook_tabs_info: list[TabInfoProtocol] | None = None
        self._front_page_frame: Optional[tk.Frame] = None
        self._toolbar = None
        self._default_notebook_page: int = 0
        self._top_level_notebook: Optional[ttk.Notebook] = None
        self._main_page_frame: Optional[tk.Frame] = None

        # Create the Tk main window
        self.mw = self._create_toplevel(title)

        # colors and fonts
        font_size = None
        if self._preferences is not None:
            font_size = self._preferences.font_size()
        self.colours, self.fonts = set_default_fonts_and_colours(self.mw,
                                                                 font_size=font_size,
                                                                invert=self.dark_mode)
        set_system_colours(self.mw, self.colours, self.theme)

        # set the filename so that it can be bound later
        self._status_bar_fall_file_info: tk.StringVar = tk.StringVar(value="None")

        # set the filename so that it can be bound later
        self._status_bar_winter_file_info: tk.StringVar = tk.StringVar(value="None")

        # set the dirty text so it can be bound later
        self._status_bar_dirty: tk.StringVar = tk.StringVar(value="")

    # ===================================================================================
    # properties
    # ===================================================================================
    @property
    def status_bar_fall_file_info(self):
        return self._status_bar_fall_file_info.get()

    @status_bar_fall_file_info.setter
    def status_bar_fall_file_info(self, value: str):
        self._status_bar_fall_file_info.set(value)

    @property
    def status_bar_winter_file_info(self):
        return self._status_bar_winter_file_info.get()

    @status_bar_winter_file_info.setter
    def status_bar_winter_file_info(self, value: str):
        self._status_bar_winter_file_info.set(value)

    @property
    def dirty_text(self):
        return self._status_bar_dirty.get()

    @dirty_text.setter
    def dirty_text(self, value):
        self._status_bar_dirty.set(value)

    # ===================================================================================
    # menu and toolbars
    # ===================================================================================
    def create_menu_and_toolbars(self, buttons: list[str], toolbar_info: dict[str:ToolbarItem],
                                 menu_details: list[MenuItem]):
        """
        Using the info, create menu and toolbar for this app
        :param buttons: a list of buttons for the toolbar
        :param toolbar_info: information about buttons/callbacks for the toolbar
        :param menu_details: all the details required for the menu
        """

        menu_bar = tk.Menu(self.mw)

        # create application specific menu
        generate_menu(self.mw, menu_details, menu_bar)
        self.mw.configure(menu=menu_bar)

        # create the auto-save part of menu
        auto_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="Auto Save", menu=auto_menu)

        tk_auto_save = tk.BooleanVar(value=bool(self._preferences.auto_save()))
        auto_menu.add_radiobutton(label="Auto Save ON", value=True, variable=tk_auto_save,
                                  command=partial(self.toggle_auto_save, True))
        auto_menu.add_radiobutton(label="Auto Save OFF", value=False, variable=tk_auto_save,
                                  command=partial(self.toggle_auto_save, False))

        # create preferences menu
        preference_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="Preferences", menu=preference_menu)
        preference_menu.add_command(label="Font Size",
                                    command=lambda: ChangeFont(self.mw, self._preferences))

        theme_menu = tk.Menu(preference_menu, tearoff=False)
        s = ttk.Style(self.mw)
        dark = 'dark' if self._preferences.dark_mode() else 'light'
        preference_menu.add_cascade(label=f"Themes ({s.theme_use()} {dark})", menu=theme_menu)
        themes = list(set(self._preferences.available_themes()).intersection(set(s.theme_names())))
        themes.sort()
        for theme in themes:
            for dark in ("light","dark"):
                theme_menu.add_command(label=f"{theme} {dark}",
                                        command=partial(self.change_theme, theme, dark))

        # create _toolbar
        self._toolbar = make_toolbar(self.mw, buttons, toolbar_info)
        self._toolbar.pack(side='top', expand=0, fill='x')

    # ===================================================================================
    # Change Theme
    # ===================================================================================
    def change_theme(self, theme, dark):
        """change the theme of the Tk"""
        self._preferences.theme(theme)
        self._preferences.dark_mode(True if dark == "dark" else False)
        self.show_message("Changing themes", msg="Changing themes will not take affect until "
                                                 "you close and re-open application")
        self._preferences.save()

    # ===================================================================================
    # status bar
    # ===================================================================================
    def create_status_bar(self):
        """Create a status bar for current filename and dirty flag"""

        mw = self.mw

        # draw frame and labels for current filename and dirty flag
        status_frame = tk.Frame(mw, borderwidth=0, relief='flat')
        status_frame.pack(side='bottom', expand=0, fill='x')

        tk.Label(status_frame, textvariable=self._status_bar_fall_file_info, borderwidth=1, relief='ridge',
              anchor='w',
              ).pack(side='left', expand=1, fill='x')

        tk.Label(status_frame, textvariable=self._status_bar_winter_file_info, borderwidth=1, relief='ridge',
              anchor='w',
              ).pack(side='left', expand=1, fill='x')

        tk.Label(status_frame, textvariable=self._status_bar_dirty, borderwidth=1, relief='ridge', width=15,
              foreground=self.colours.DirtyColour).pack(side='right', fill='x')

    # ===================================================================================
    # welcome page
    # ===================================================================================
    def create_welcome_page_base(self, logo: str) -> tk.Frame:
        """
        Creates the very first page that is shown to the user
        :param logo: image to use as part of the front page header
        :return: the frame for the super class to use to add whatever they want
        """
        mw = self.mw
        self._front_page_frame = tk.Frame(mw, borderwidth=10, relief='flat', background=self.colours.DataBackground)
        self._front_page_frame.pack(side='top', expand=1, fill='both')

        # create an image object of the _logo
        # ... for some weird reason, if logo is not part of 'self', then it doesn't work
        # Weird, right?
        self.logo = tk.PhotoImage(file=logo)

        # frame and label
        tk.Label(self._front_page_frame, image=self.logo, borderwidth=0, relief='flat') \
            .pack(side='left', expand=0)

        # --------------------------------------------------------------
        # frame for holding buttons for starting the scheduling/allocation tasks
        # --------------------------------------------------------------
        option_frame = tk.Frame(
            self._front_page_frame,
            background=self.colours.DataBackground,
            borderwidth=10,
            relief='flat')
        option_frame.pack(side='left', expand=1, fill='both')

        tk.Frame(option_frame, background=self.colours.DataBackground).pack(expand=1, fill='both')
        center_frame = tk.Frame(option_frame, background=self.colours.DataBackground)
        center_frame.pack(expand=0, fill='both')
        tk.Frame(option_frame, background=self.colours.DataBackground).pack(expand=1, fill='both')

        return center_frame

    # ===================================================================================
    # standard page with notebook
    # ===================================================================================
    def create_standard_page(self, notebook_pages_info: Optional[list[TabInfoProtocol]] = None, reset=False):
        """
        Create the 'welcome' page after the main page has fulfilled its purpose
        :param notebook_pages_info: which tabs, and callbacks, etc are in the notebook being created
        :param reset: should the notebook be re-drawn
        """

        # if the page is already created, do not recreate it
        if self._notebook_frame is not None and not reset:
            if self._notebook_frame.main_notebook_frame:
                self._notebook_frame.main_notebook_frame.select(self._default_notebook_page)
                self._notebook_frame.main_notebook_frame.event_generate("<<NotebookTabChanged>>")
                return

        # create the page
        mw = self.mw
        self._front_page_frame.destroy()

        # frame
        if self._main_page_frame is not None:
            self._main_page_frame.destroy()
        self._main_page_frame = tk.Frame(mw, borderwidth=1, relief='ridge')
        self._main_page_frame.pack(side='top', expand=1, fill='both')
        self._notebook_frame = NoteBookFrameTk(self.mw, self._main_page_frame, notebook_pages_info,
                                               self.weird_mac_crap_tab_change_handler)

    def _exit_schedule(self, *_):
        self.mw.destroy()

    def weird_mac_crap_tab_change_handler(self, name, frame):
        self.recursive_expose(frame)
        self.notebook_tab_changed_handler(name, frame)

    def recursive_expose(self, widget):
        for w in widget.winfo_children():
            self.recursive_expose(w)
            widget.update()

    def select_tab(self, name):
        for notebook in self._notebook_frame.notebooks:
            tab_names = {notebook.tab(i, option="text"):i for i in notebook.tabs()}
            if name in tab_names.keys():
                notebook.select(tab_names[name])
                self._notebook_frame.tab_changed(notebook,tab_names[name] )

    # ========================================================================
    # choose file
    # ========================================================================
    def select_file_to_save(self, title:str="Save File As") -> str:
        return self._select_file(partial(tk_fd.asksaveasfilename,
                                         defaultextension="csv",
                                         title=title))

    def select_file_to_open(self, title = "Open Schedule") -> str:
        return self._select_file(partial(tk_fd.askopenfilename,
                                         title=title))

    def _select_file(self, select_gui) -> str:
        current_dir = self._preferences.current_dir() or self._preferences.home_directory()
        kwargs = {"filetypes": (
            ('schedule files', '*.csv'),
            ('All files', '*.*')
        )}

        if current_dir is not None:
            kwargs["initialdir"] = current_dir

        filename = select_gui(**kwargs)
        return filename

    # ========================================================================
    # message boxes
    # ========================================================================
    def show_error(self, title: str, msg: str, detail: str = ""):
        showerror(title=title, message=msg, detail=detail, icon='error')

    def show_message(self, title: str, msg: str, detail: str = ""):
        showinfo(title=title, message=msg, detail=detail, icon='info')

    def ask_yes_no(self, title: str, msg: str, detail: str = ""):
        return askyesno(title=title, message=msg, detail=detail, icon='info')

    def show_custom_message(self, title="", msg=""):
        dialog = tk.Tk()
        dialog.title(title)

        label = tk.Label(dialog, text=msg, justify='left')
        label.pack(pady=20,padx=20)

        ok_button = tk.Button(dialog, text="OK", command=dialog.destroy)
        ok_button.pack(padx=20,pady=20, expand=1)

        dialog.transient(self.mw)  # Make it a child of the main window
        dialog.grab_set()      # Make it modal
        self.mw.wait_window(dialog) # Wait for the dialog to close

    # ===================================================================================
    # create top level
    # ===================================================================================
    def _create_toplevel(self, title):
        """Create the top level window, specify fonts and colors"""
        # create main window and frames
        """Create the top level window, specify fonts and colors"""
        # create main window and frames
        mw = tk.Tk()
        mw.title(title)
        mw.geometry(f"{WELCOME_WIDTH}x{WELCOME_HEIGHT}")
        #mw.geometry("100x100")

        # when clicking the 'x' in the corner of the window, call _exit_schedule
        mw.protocol("WM_DELETE_WINDOW", self._exit_schedule)

        # have main window pop-up and have focus
        # https://stackoverflow.com/questions/8691655/how-to-put-a-tkinter-window-on-top-of-the-others/8691795
        mw.lift()
        mw.call('wm', 'attributes', '.', '-topmost', True)
        mw.after_idle(mw.call, 'wm', 'attributes', '.', '-topmost', False)
        mw.focus_force()

        return mw

    # ===================================================================================
    # start event loop
    # ===================================================================================
    def start_event_loop(self):
        """time_start the Tk event main event loop"""
        self.mw.mainloop()

    # ===================================================================================
    # waiting routines
    # ===================================================================================
    def wait_for_it(self, title, msg):
        self.stop_waiting()

        self._wait = self.mw.winfo_toplevel()
        self._wait.title = title
        tk.Label(self._wait, text=msg).pack(expand=1, fill='both')

        self._wait.geometry('300x450')
        self.mw.update()

    def stop_waiting(self):
        if self._wait:
            self._wait.destroy()
        self._wait = None



"""Create a scrollable frame, although it doesn't attach any scrollbars"""

from __future__ import annotations
import tkinter as tk

import sys
import traceback
from tkinter import ttk
from typing import Literal, Optional


def eprint(*args, **kwargs):
    print("\n", file=sys.stderr, **kwargs)
    print(*args, file=sys.stderr, **kwargs)
    print(traceback.format_stack()[0], file=sys.stderr, **kwargs)
    exit()


###############
# https://docstore.mik.ua/orelly/perl3/tk/ch06_03.htm
# Section 6.3.8
# http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/events.html
# https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/index.html
###############

class Pane(tk.Frame):
    """
        ┌──────────────────────────────────────────────────────┐
        │ ┌──────────────────────────────────────────────────┐ │
        │ │ ┌──────────────────────────────────────────────┐ │ │
        │ │ │                                              │ │ │
        │ │ │                                              │ │ │
        │ │ │                                              │< ------- inner frame
        │ │ │             Inner frame                      │ │ │
        │ │ │          this is where you add stuff         │ │<------ scrollable canvas
        │ │ │                                              │ │ │
        │ │ └──────────────────────────────────────────────┘ │ │<---- self (Pane)
        │ └──────────────────────────────────────────────────┘ │
        └──────────────────────────────────────────────────────┘

    """

    # ==============================================================================================
    # properties
    # ==============================================================================================
    @property
    def frame(self)->tk.Widget:
        return self._widget

    @property
    def canvas(self):
        return self._scrollable_object

    # ==============================================================================================
    # constructor
    # ==============================================================================================
    def __init__(self, parent, **kwargs):
        background_colour = None
        if "bg" in kwargs:
            background_colour = kwargs['bg']
        if "background" in kwargs:
            background_colour = kwargs['background']

        tk.Frame.__init__(self, parent)

        self.horizontal_scrollbar: Optional[ttk.Scrollbar] = None
        self.vertical_scrollbar: Optional[ttk.Scrollbar] = None

        canvas = tk.Canvas(self, bg="red")
        canvas.pack(side="left", fill="both", expand=tk.TRUE)
        self._scrollable_object = canvas
        super().configure(background=background_colour)
        super().configure(highlightbackground=background_colour)
        canvas.configure(background=background_colour)
        canvas.configure(highlightbackground=background_colour)

        # Create a frame inside the canvas which will be scrolled with it.
        self._widget = tk.Frame(canvas, **kwargs)
        self._scrolled_id = canvas.create_window(0, 0, window=self._widget,
                                           anchor="nw")

        self._widget.bind('<Configure>', self.configure_interior)

    def configure_interior(self, *_):
        # Update canvas to match the size of the inner frame.
        width, height = (self._widget.winfo_reqwidth(), self._widget.winfo_reqheight())
        self.canvas.config(scrollregion=(0, 0, width, height), width=width, height=height)

    # ===============================================================================================================
    # pass on all 'configure' to the scrollable object
    # ===============================================================================================================
    def configure(self, **kwargs):
       self._widget.configure(**kwargs)

    def yview(self, *args):
        self.canvas.yview(*args)

    def xview(self, *args):
        self.canvas.xview(*args)

    # ===============================================================================================================
    # scrolling methods
    # ===============================================================================================================
    def see(self, widget, **_kwargs)->tuple[Optional[float], Optional[float]]:
        """Adjusts the view so that widget is visible.

        Additional parameters in options-value pairs can be passed,
        each option-value pair must be one of the following

        NOT IMPLEMENTED YET

        -anchor => anchor
        Specifies how to make the widget visible. If not given then as much of the widget as possible is made visible.

        Possible values are n, s, w, e, nw, ne, sw and se.
        This will cause an edge on the widget to be aligned with the corresponding edge on the pane.
        for example nw will cause the top left of the widget to be placed at the top left of the pane.
        s will cause the bottom of the widget to be placed at the bottom of the pane, and as much of
        the widget as possible made visible in the x direction.

        """
        delta_x = self.xview_widget(widget)
        delta_y = self.yview_widget(widget)
        return delta_x, delta_y

    def yview_moveto(self, fraction: Optional[float]):
        """
            Adjusts the view in the window so that fraction of the total width of the Scrollable object is off-screen
            to the top.

            fraction must be a fraction between 0 and 1.
        """
        if fraction is not None:
            if self.vertical_scrollbar is not None:
                self.canvas.yview_moveto(fraction)

    def xview_moveto(self, fraction: Optional[float]):
        """
            Adjusts the view in the window so that fraction of the total width of the Scrollable object is off-screen
            to the left.

            fraction must be a fraction between 0 and 1.
        """
        if fraction is not None:
            if self.horizontal_scrollbar is not None:
                self.canvas.xview_moveto(fraction)

    def xview_widget(self, widget=None) -> Optional[float]:
        """
        No parameters:
            Returns a list containing two elements, both of which are real fractions between 0 and 1.
            The first element gives the position of the left of the window, relative to the scrollable object as a whole
            (0.5 means it is halfway through the Frame, for example).
            The second element gives the position of the right of the window, relative to the scrollable object
            as a whole.

        Tk Widget defined
            Adjusts the view in the window so that widget is displayed.

         """

        # make sure idle tasks are finished (includes redrawing of changes in geometry)
        self.update_idletasks()
        pos: Optional[float] = None

        if self.horizontal_scrollbar is None:
            return pos

        if widget is not None:
            scrollable_object_width = self._widget.winfo_width()
            scrollable_object_left = self.frame.winfo_rootx()

            to_be_seen_left = widget.winfo_rootx()
            to_be_seen_width = widget.winfo_width()
            to_be_seen_right = to_be_seen_left + to_be_seen_width

            scroll_region_width = self._scrollable_object.winfo_width()
            scroll_region_left = self._scrollable_object.winfo_rootx()
            scroll_region_right = scroll_region_left + scroll_region_width

            if to_be_seen_left < scroll_region_left:
                dx = to_be_seen_left - scroll_region_left
                pos = (scroll_region_left - scrollable_object_left + dx) / scrollable_object_width
                self._scrollable_object.xview_moveto(pos)

            if to_be_seen_right > scroll_region_right:
                dx = to_be_seen_right - scroll_region_right + to_be_seen_width
                pos = (scroll_region_left - scrollable_object_left + dx) / scrollable_object_width
                self._scrollable_object.xview_moveto(pos)

        return pos

    def yview_widget(self, widget=None) -> Optional[float]:
        """
        No parameters:
            Returns a list containing two elements, both of which are real fractions between 0 and 1.
            The first element gives the position of the top of the window, relative to the scrollable object as a whole
            (0.5 means it is halfway through the Frame, for example).
            The second element gives the position of the bottom of the window, relative to the scrollable object
            as a whole.

        Tk Widget defined
            Adjusts the view in the window so that widget is displayed.

        """

        # make sure idle tasks are finished (includes redrawing of changes in geometry)
        self.update_idletasks()
        pos = None

        if self.vertical_scrollbar is None:
            return pos

        if widget is not None:
            scrollable_object_height = self._widget.winfo_height()
            scrollable_object_top = self.frame.winfo_rooty()

            to_be_seen_top = widget.winfo_rooty()
            to_be_seen_height = widget.winfo_height()
            to_be_seen_bottom = to_be_seen_top + to_be_seen_height

            scroll_region_height = self._scrollable_object.winfo_height()
            scroll_region_top = self._scrollable_object.winfo_rooty()
            scroll_region_bottom = scroll_region_top + scroll_region_height

            if to_be_seen_top < scroll_region_top:
                dy = to_be_seen_top - scroll_region_top
                pos = (scroll_region_top - scrollable_object_top + dy) / scrollable_object_height
                self._scrollable_object.yview_moveto(pos)

            if to_be_seen_bottom > scroll_region_bottom:
                dy = to_be_seen_bottom - scroll_region_bottom + to_be_seen_height
                pos = (scroll_region_top - scrollable_object_top + dy) / scrollable_object_height
                self._scrollable_object.yview_moveto(pos)

        return pos

    def xview_scroll(self, number: int, what: Literal["units", "pages"]) -> None:
        """ Shift the x-view according to NUMBER which is measured in "units" or "pages" (WHAT).

        'what' must be  Literal["units", "pages"]
        """
        if self.horizontal_scrollbar is not None:
            self._scrollable_object.xview_scroll(number, what)

    def yview_scroll(self, number: int, what: Literal["units", "pages"]) -> None:
        """ Shift the y-view according to NUMBER which is measured in "units" or "pages" (WHAT).

        'what' must be  Literal["units", "pages"]
        """
        if self.vertical_scrollbar is not None:
            self._scrollable_object.yview_scroll(number, what)



from __future__ import annotations
import tkinter as tk
from tkinter import ttk

import sys
import traceback
import importlib
from typing import Any, Literal


def eprint(*args, **kwargs):
    print("\n", file=sys.stderr, **kwargs)
    print(*args, file=sys.stderr, **kwargs)
    print(traceback.format_stack()[0], file=sys.stderr, **kwargs)
    exit()


###############
# https://docstore.mik.ua/orelly/perl3/tk/ch06_03.htm
# Section 6.3.8
# http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/events.html
# https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/index.html
###############

class Scrolled(tk.Frame):
    """Creates a frame, creates and inserts the defined widget, and adds scrollbars

Parameters
----------
    parent: Tk_container
    widget_type: str ->What widget you want to create

Options
-------
    scrollbars:str -> defines which scrollbars you want and where you want them: "nsew".

                        'n' is a horizontal scrollbar at the top of the widget

                        'e' is a vertical scrollbar on the right side of the widget

                        you can only have one scrollbar for horizontal and one for vertical

    all other options will be passed to the widget when it is created


Properties
----------
    widget -> the widget that you asked to be created

    horizontal_scrollbar -> the horizontal scrollbar that was created

    vertical_scrollbar -> the vertical scrollbar that was created

Methods
-------
    see (self,internal_widget)

    yview_moveto(self, fraction)

    xview_moveto(self, fraction)

    yview(self, widget=None)

    xview(self, widget=None)

    yview_scroll(self, number, what)

    xview_scroll(self, number, what)


Notes
-----
    The created object will be a frame, so you can pack or grid it as desired

    To access the scrolled widget, use 'self.widget' property

    To access the scrollbars, use self.horizontal_scrollbar and self.vertical_scrollbar

     - scrollbars have been set_default_fonts_and_colours to not take focus, if you want to
     change that you have to modify
       the scrollbars after the scrolled widget has been created


Examples
--------

example::

        from tkinter import *
        from scrolled import Scrolled

        mw = Tk()
        scrolled_listbox = Scrolled(mw, 'Listbox', scrollbars=E, bg='pink')
        scrolled_listbox.pack(fill=BOTH, side=TOP, expand=1)
        scrolled_listbox.vertical_scrollbar.configure(width=30)

        for i in range(20):
            scrolled_listbox.widget.insert("end", f"entry {i=}")

        mw.mainloop()

    Example
        from tkinter import *
        from scrolled import Scrolled

        mw = Tk()
        scrolled_frame = Scrolled(mw, 'Frame', scrollbars=E)
        scrolled_frame.pack(fill=BOTH, side=TOP, expand=1)

        for i in range(10):
            Button(scrolled_frame.widget,text=f"Button {i}").pack(ipady=10)

        mw.mainloop()
    """

    # ==============================================================================================
    # properties
    # ==============================================================================================
    @property
    def widget(self)->tk.Widget:
        return self._widget

    @property
    def horizontal_scrollbar(self)->tk.Scrollbar:
        return self._horizontal_scrollbar

    @property
    def vertical_scrollbar(self)->tk.Scrollbar:
        return self._vertical_scrollbar

    @property
    def scroll_widget(self):
        return self._scrollable_object

    # ==============================================================================================
    # constructor
    # ==============================================================================================
    def __init__(self, parent, widget_type, scrollbars="e", **kwargs):
        """ constructs scrollbars, widget from widget_type, attaches them appropriately

        a frame will be created (self)
        a widget of widget resource_type will be created inside the frame (self.widget)
        scrollbars will be created as requested (self.horizontal_scrollbar, self.vertical_scrollbar)

        NOTE: parent frame will use 'pack', so you can't "grid" on this parent frame
        """
        # ----------------------------------------------------------------------------------------
        # initialize_columns the holding frame
        # ----------------------------------------------------------------------------------------
        tk.Frame.__init__(self, parent)

        self._widget_type = widget_type
        self._scrollable_object = None
        self._vertical_scrollbar = None
        self._horizontal_scrollbar = None
        self._widget = None

        scrolled_module = Scrolled.__module__
        parts = scrolled_module.split('.')
        modified_tk_module = importlib.import_module('.'.join(parts[:-1]))

        # ----------------------------------------------------------------------------------------
        # get the Tk object that needs to be created
        # ----------------------------------------------------------------------------------------
        _tk_widget_type = None

        # looking for a Tk widget
        try:
            _tk_widget_type = getattr(tk, widget_type)
        except AttributeError:

            # looking for an expanded tk widget from ttk
            try:
                _tk_widget_type = getattr(ttk, widget_type)
            except AttributeError:

                # looking for a personalized widget in schedule.Tk
                try:
                    _tk_widget_type = getattr(modified_tk_module, widget_type)
                except AttributeError as e:
                    eprint(e)

        # ----------------------------------------------------------------------------------------
        # bail out if user is trying to create 2 scrollbars in the same orientation
        # ----------------------------------------------------------------------------------------
        if ('e' in scrollbars and 'w' in scrollbars) or ('n' in scrollbars and 's' in scrollbars):
            eprint(f"You cannot have two horizontal or two vertical scrollbars ({scrollbars=})")

        # ----------------------------------------------------------------------------------------
        # create scrollbars
        # ----------------------------------------------------------------------------------------
        self._horizontal_scrollbar = None
        self._vertical_scrollbar = None
        if 'n' in scrollbars or 's' in scrollbars:
            self._horizontal_scrollbar = ttk.Scrollbar(self, orient='horizontal', takefocus=0)
            if 'n' in scrollbars:
                self._horizontal_scrollbar.pack(side='top', fill='x')
            else:
                self._horizontal_scrollbar.pack(side='bottom', fill='x')

        if 'e' in scrollbars or 'w' in scrollbars:
            self._vertical_scrollbar = ttk.Scrollbar(self, orient='vertical', takefocus=0)
            if 'e' in scrollbars:
                self._vertical_scrollbar.pack(side='right', fill='y')
            else:
                self._vertical_scrollbar.pack(side='left', fill='y')

        # ----------------------------------------------------------------------------------------
        # is the widget we want to scroll, scrollable?
        # ----------------------------------------------------------------------------------------
        try:
            _xview = getattr(_tk_widget_type, "xview")
            _yview = getattr(_tk_widget_type, "yview")

            self._widget = _tk_widget_type(self, **kwargs)
            self._scrollable_object = self._widget
            self._scrollable_object.pack(fill='both', expand=1)


        except AttributeError:
            pass
            self._canvas_scrolled(**kwargs)

        # ----------------------------------------------------------------------------------------
        # bind the scrollable object to the scrollbars
        # ----------------------------------------------------------------------------------------

        if self._vertical_scrollbar is not None:
            self._vertical_scrollbar.configure(command=self._scrollable_object.yview)
            self._scrollable_object.configure(yscrollcommand=self._vertical_scrollbar.set)

        if self._horizontal_scrollbar is not None:
            self._horizontal_scrollbar.configure(command=self._scrollable_object.xview)
            self._scrollable_object.configure(xscrollcommand=self._horizontal_scrollbar.set)

        self.pack(expand=1, fill="both")


    def _canvas_scrolled(self, **kwargs):
        # ----------------------------------------------------------------------------------------
        # the widget is not scrollable, so we are going to put it inside a canvas object
        # which is scrollable
        # ----------------------------------------------------------------------------------------
        # Create a canvas object and a vertical scrollbar for scrolling it.
        canvas = tk.Canvas(self)
        canvas.pack(side="left", fill="both", expand=tk.TRUE)
        self._scrollable_object = canvas

        # Create a frame inside the canvas which will be scrolled with it.
        self._widget = tk.Frame(canvas)

        self._scrolled_id = canvas.create_window(0, 0, window=self._widget,
                                           anchor="nw")

        # Track changes to the canvas and frame width and sync them,
        # also updating the scrollbar.
        def _configure_interior(event):
            # Update the scrollbars to match the size of the inner frame.
            size = (self._widget.winfo_reqwidth(), self._widget.winfo_reqheight())
            canvas.config(scrollregion=(0,0, *size))

        self._widget.bind('<Configure>', _configure_interior)

    # ===============================================================================================================
    # pass on all 'configure' to the scrollable object
    # ===============================================================================================================
    def configure(self, **kwargs):
       self._widget.configure(**kwargs)

    # ===============================================================================================================
    # Subwidget
    # ===============================================================================================================
    def Subwidget(self, name: str) -> Any:
        """returns scrollbar, the widget to be scrolled, or the scrolled object"""
        if name == 'xscrollbar':
            return self._horizontal_scrollbar
        if name == 'yscrollbar':
            return self._vertical_scrollbar
        if name == self._widget_type:
            return self._widget
        if name == 'scrollable':
            return self._scrollable_object

    # ===============================================================================================================
    # scrolling methods
    # ===============================================================================================================
    def see(self, widget, **_kwargs):
        """Adjusts the view so that widget is visible.

        Additional parameters in options-value pairs can be passed,
        each option-value pair must be one of the following

        NOT IMPLEMENTED YET

        -anchor => anchor
        Specifies how to make the widget visible. If not given then as much of the widget as possible is made visable.

        Possible values are n, s, w, e, nw, ne, sw and se.
        This will cause an edge on the widget to be aligned with the corresponding edge on the pane.
        for example nw will cause the top left of the widget to be placed at the top left of the pane.
        s will cause the bottom of the widget to be placed at the bottom of the pane, and as much of
        the widget as possible made visible in the x direction.

        """

        self.xview_widget(widget)
        self.yview_widget(widget)

    def yview_moveto(self, fraction):
        """
            Adjusts the view in the window so that fraction of the total width of the Scrolloable object is off-screen
            to the top.

            fraction must be a fraction between 0 and 1.
        """
        if self._vertical_scrollbar is not None:
            self._scrollable_object.yview_moveto(fraction)

    def xview_moveto(self, fraction):
        """
            Adjusts the view in the window so that fraction of the total width of the Scrolloable object is off-screen
            to the left.

            fraction must be a fraction between 0 and 1.
        """
        if self._horizontal_scrollbar is not None:
            self._scrollable_object.xview_moveto(fraction)

    def xview_widget(self, widget=None):
        """
        No parameters:
            Returns a list containing two elements, both of which are real fractions between 0 and 1.
            The first element gives the position of the left of the window, relative to the scrollable object as a whole
            (0.5 means it is halfway through the Frame, for example).
            The second element gives the position of the right of the window, relative to the scrollable object
            as a whole.

        Tk Widget defined
            Adjusts the view in the window so that widget is displayed.

         """

        # make sure idle tasks are finished (includes redrawing of changes in geometry)
        self.update_idletasks()

        if self.horizontal_scrollbar is None:
            return 0, 1

        if widget is not None:
            # NOT IMPLEMENTED YET
            pass

    def yview_widget(self, widget=None):
        """
        No parameters:
            Returns a list containing two elements, both of which are real fractions between 0 and 1.
            The first element gives the position of the top of the window, relative to the scrollable object as a whole
            (0.5 means it is halfway through the Frame, for example).
            The second element gives the position of the bottom of the window, relative to the scrollable object
            as a whole.

        Tk Widget defined
            Adjusts the view in the window so that widget is displayed.

        """

        # make sure idle tasks are finished (includes redrawing of changes in geometry)
        self.update_idletasks()

        if self.vertical_scrollbar is None:
            return 0, 1  # no scrollbar

        #print("yview:", *args, **kwargs)
        #return self._scrollable_object.xview(*args, **kwargs)
        if widget is not None:
            scrollable_object_height = self._widget.winfo_height()
            scrollable_object_top = self.widget.winfo_rooty()

            to_be_seen_top = widget.winfo_rooty()
            to_be_seen_height = widget.winfo_height()
            to_be_seen_bottom = to_be_seen_top + to_be_seen_height

            scroll_region_height = self._scrollable_object.winfo_height()
            scroll_region_top = self._scrollable_object.winfo_rooty()
            scroll_region_bottom = scroll_region_top + scroll_region_height

            if to_be_seen_top < scroll_region_top:
                dy = to_be_seen_top - scroll_region_top
                pos = (scroll_region_top - scrollable_object_top + dy) / scrollable_object_height
                self._scrollable_object.yview_moveto(pos)

            if to_be_seen_bottom > scroll_region_bottom:
                dy = to_be_seen_bottom - scroll_region_bottom + to_be_seen_height
                pos = (scroll_region_top - scrollable_object_top + dy) / scrollable_object_height
                self._scrollable_object.yview_moveto(pos)

        return self.vertical_scrollbar.get()

    def xview_scroll(self, number: int, what: Literal["units", "pages"]) -> None:
        """ Shift the x-view according to NUMBER which is measured in "units" or "pages" (WHAT).

        'what' must be  Literal["units", "pages"]
        """
        if self.horizontal_scrollbar is not None:
            self._scrollable_object.xview_scroll(number, what)

    def yview_scroll(self, number: int, what: Literal["units", "pages"]) -> None:
        """ Shift the y-view according to NUMBER which is measured in "units" or "pages" (WHAT).

        'what' must be  Literal["units", "pages"]
        """
        if self.vertical_scrollbar is not None:
            self._scrollable_object.yview_scroll(number, what)



"""
Change the default font size for the application
"""
from __future__ import annotations

import tkinter as tk
from tkinter.simpledialog import Dialog
from tkinter.messagebox import showinfo
from ..Utilities.Preferences import Preferences
from ..modified_tk import TkFonts


class ChangeFont(Dialog):
    def __init__(self, mw: tk.Tk, preferences: Preferences, title="Change Font"):
        self.warning_label = None
        self.label_big = None
        self.label_normal = None
        self.label_small = None
        self.preferences = preferences
        self.font_size_tk = tk.StringVar(value=str(self.preferences.font_size()))
        self.fonts = TkFonts(mw, self.preferences.font_size())

        super().__init__(mw, title)

    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame:tk.Frame):
        """
        :param frame: the frame where you are gonna put stuff
        """
        subframe = tk.Frame(frame)
        self.frame=frame

        self.label_small:tk.Label = tk.Label(frame, text="Small text: ABC abc 0123", font=self.fonts.small)
        self.label_normal:tk.Label = tk.Label(frame, text="Normal text: ABC abc 0123", font=self.fonts.normal)
        self.label_big:tk.Label = tk.Label(frame, text="Big text: ABC abc 0123", font=self.fonts.big)
        self.warning_label = tk.Label(frame, text="Must restart application for font changes to take effect!!",
                 font=self.fonts.bigbold)

        modify_font = frame.winfo_toplevel().register(self._change_font)
        font_size_label=tk.Label(subframe, text="Font Size:")
        font_size_entry = tk.Entry(subframe, textvariable=self.font_size_tk,
                                   validate='key',
                                   validatecommand=(modify_font, '%P', '%s')
                                   )

        # layout
        self.warning_label.pack(side="top", expand=1, fill='x', pady=20, padx=10)
        subframe.pack(side="top", expand=1, fill='x', pady=20)
        font_size_label.pack(side='left', expand=1, fill='x', padx=10)
        font_size_entry.pack(side='left', expand=1, fill='x', padx=10)

        self.label_small.pack(side='top', expand=1, fill='x', padx=20)
        self.label_normal.pack(side='top', expand=1, fill='x', padx=20)
        self.label_big.pack(side='top', expand=1, fill='x', padx=20)

        return font_size_entry

    # ================================================================================================================
    # change the displayed font size
    # ================================================================================================================
    def _change_font(self, number: str, *_):
        if number == "":
            return True
        try:
            int(number)
            self.fonts = TkFonts(self.frame.winfo_toplevel(), int(number))
            self.warning_label.config(font=self.fonts.bigbold)
            self.label_big.config(font=self.fonts.big)
            self.label_small.config(font=self.fonts.small)
            self.label_normal.config(font=self.fonts.normal)
            return True
        except ValueError:
            return False

    # ================================================================================================================
    # validate before applying
    # ================================================================================================================
    def validate(self):
        """
        Is the data, as entered by the user, valid?
        :return: True if data is good (changes are applied), false otherwise (nothing happens)
        """
        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        """apply the changes and close the dialog"""
        try:
            font_size = int(self.font_size_tk.get())
            self.preferences.font_size(font_size)
            self.preferences.save()
            showinfo("Changing themes", "Changing fonts will not take affect until "
                                                     "you close and re-open application")

        except ValueError:
            pass



"""
DESCRIPTION
Creates and keep tracks of IDs for collections of things

SYNOPSIS

class Foo
    ids = id_gen.IdGenerator()

    def __init__(self, id: Optional[int])
        self.id = Foo.ids.get_new_id(block_id)
"""

from typing import Optional


class IdGenerator:
    """
    A generic class to create and keep track of id numbers for collections of objects
    """

    def __init__(self, the_id: int = -1):
        self._current_id = the_id
        self._iter = self.__iter__()

    def __iter__(self):
        while True:
            self._current_id += 1
            yield self._current_id

    # def _next_id(self, this_id: Optional[int] = None) -> int:
    #     if this_id is not None and this_id > self._current_id:
    #         self._current_id = this_id - 1
    #     return next(self._iter)

    @property
    def current_id(self) -> int:
        """Value of the largest id that was created"""
        return self._current_id

    def get_new_id(self, value: Optional[int|str] = None) -> int:
        """Change the current id, else get the next value"""

        if value is None:
            value = next(self._iter)
        else:
            try:
                value = int(value)
            except TypeError:
                value = next(self._iter)

        self._current_id = max(self._current_id, value)
        return value



""" SYNOPSIS/EXAMPLE:
    tbd
"""

from __future__ import annotations
from .enums import WeekDay

MINUTE_BLOCK_SIZE = 30
MINIMUM_DURATION = 0.5
DEFAULT_DAY = WeekDay.Monday
DEFAULT_START = 8.0
DEFAULT_HOURS = 8
DEFAULT_DURATION = 1.5
MIN_START_TIME = 8
MAX_END_TIME = 18
MAXIMUM_DURATION = 8

def get_hour_minutes_from_hours(hours: float) -> (int, int):
    """converts number of hours (as a float) to integer hour and integer minutes"""
    hour = int(hours)
    minute = (hours % 1) * 60
    return hour, int(minute)

def get_clock_string_from_hours(hours: float)->str:
    hour,minute = get_hour_minutes_from_hours(hours)
    return f"{hour}:{minute:02d}"

# ============================================================================
# TimeSlot
# ============================================================================
class TimeSlot:
    """
    A time slot is specified by a day of the week, time_start time, length (in hours), and whether
        it is allowed to move.
    """

    # ------------------------------------------------------------------------
    # constructor
    # ------------------------------------------------------------------------
    def __init__(self, day: WeekDay | float = DEFAULT_DAY ,
                 start: float = DEFAULT_START,
                 duration: float = DEFAULT_DURATION,
                 movable: bool = True):
        """
        :param day: Day of the week expressed as a string or as a Weekday
        :param start: Start time using the 24hr clock (i.e., 1PM is "13:00")
        :param duration: How long this class lasts, in hours.
        :param movable: Whether this time_slot can be moved or not.
        """
        if isinstance(day, float) or isinstance(day, int):
            day = WeekDay(round(day))
        self.day: WeekDay = day
        self.start: float = start
        self.duration: float = min(max(duration, MINIMUM_DURATION), MAXIMUM_DURATION)
        self.movable: bool = movable

    # ------------------------------------------------------------------------
    # properties
    # ------------------------------------------------------------------------
    @property
    def end(self) -> float:
        return self.start + self.duration

    # ------------------------------------------------------------------------
    # snap_to_time
    # ------------------------------------------------------------------------
    def snap_to_time(self):

        # it's not movable!
        if not self.movable:
            return

        hour_fraction = int(60/MINUTE_BLOCK_SIZE)

        # duration needs to be snapped to the same time span
        duration = max(MINIMUM_DURATION, round(self.duration * hour_fraction) / hour_fraction)
        self.duration = duration

        # snap to the number of minutes in the block size
        start = max(self.start, MIN_START_TIME)
        if start + duration > MAX_END_TIME:
            start = MAX_END_TIME - duration
        hour, minute = get_hour_minutes_from_hours(start)
        minute = round(minute / MINUTE_BLOCK_SIZE) * MINUTE_BLOCK_SIZE

        # update scheduled_time
        self.start = hour + minute/60


    # ------------------------------------------------------------------------
    # snap_to_day
    # ------------------------------------------------------------------------
    def snap_to_day(self, fractional_day: float,
                    min_day: WeekDay = WeekDay.Monday,
                    max_day: WeekDay = WeekDay.Friday) -> bool:
        """
        :param fractional_day: an FLOAT that indicates where the day should be
        :param min_day: first day of the workweek
        :param max_day: last day of the workweek
        :return: True if modifications were made
        """
        day = round(fractional_day)
        day = max(min(day, max_day.value), min_day.value)
        if self.day.value == day:
            return False
        self.day = WeekDay(day)
        return True

    # ------------------------------------------------------------------------
    # conflicts
    # ------------------------------------------------------------------------
    def conflicts_time(self, other: TimeSlot, delta: float = 0.05) -> bool:
        """
        Tests if the current Time_Slot conflicts with another TimeSlot.
        :param other: other timeslot
        :param delta: the amount of leeway that we are allowing for in floating pt arithmetic
        """

        # detect date collisions.
        if self.day != other.day:
            return False

        # Calculate the time_start/end for each blocks within error factor.
        self_start = self.start + delta
        self_end = self.start + self.duration - delta
        other_start = other.start + delta
        other_end = other.start + other.duration - delta

        conflict1 = other_start < self_start < other_end or other_start < self_end < other_end
        conflict2 = self_start >= other_start and self_end <= other_end
        conflict3 = other_start >= self_start and other_end <= self_end
        return conflict1 or conflict2 or conflict3

    # ------------------------------------------------------------------------
    # sorting, string representation, etc
    # ------------------------------------------------------------------------
    def __eq__(self, other):
        return (self.day == other.day
                and self.start == other.start
                and self.duration == other.duration)

    def __lt__(self, other):
        return ((self.day, self.start, self.duration) <
                (other.day, other.start, other.duration))

    def __str__(self):
        return (f"{self.day.name}: {get_clock_string_from_hours(self.start)} "
                f"({self.duration} hrs)")

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))



import tkinter as tk
from functools import partial
from tkinter import ttk
import tkinter.messagebox as message_box

from ..gui_generics.number_validations import entry_float
from ..model import WeekDay


# ================================================================================================================
# class times?
# ================================================================================================================
def validate_class_times_equals_course_time(block_tk_variables, course_hours: float)-> bool:
    """
    Compares the total duration of all the blocks and compares to the hours assigned to the course
    :param block_tk_variables: list of tuples with string reprs of day, start_time, and duration
    :param course_hours: the hours assigned to this course
    :return: true if no blocks defined, or if duration is equal to assigned hours, or user oks the discrepancy
    """
    total = 0.0
    for block_info in block_tk_variables:
        duration = float(block_info[2].get().strip().split(" ")[0])
        if duration < 0:
            duration += 24
        total += duration

    if total == course_hours or total == 0.0:
        return True

    else:
        return message_box.askyesno("Class Times", f"Total allocated class times ({total})\n"
                                       f"does not equal course time ({course_hours})"
                                       f"\n\nDo you wish to continue?")


# ================================================================================================================
# get updated block information
# ================================================================================================================
def get_block_info_from_tk_widgets(block_tk_variables: list[tk.StringVar]) -> list[tuple[float,float,float]]:
    """
    Convert string (day, start_time, duration) to int(day), float(start_time), float(duration)
    :param block_tk_variables:
    :return: updated list
    """
    updated_blocks = []
    for i, b in enumerate(block_tk_variables):
        d = [x.get() for x in b]
        day = WeekDay[d[0]].value
        start = float(d[1].strip())
        duration = float(d[2].strip())
        updated_blocks.append((day,start,duration))

    return updated_blocks

# ================================================================================================================
# update the gui with block data
# ================================================================================================================
def refresh_gui_blocks(self,):
    """
    Create the drop-down and entry boxes etc for blocks
    :param self: the object that contains all the block info, frame, etc
    """
    days = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]

    # remove any pre-existing stuff
    for w in self.block_frames.winfo_children():
        w.destroy()

    row_frames = []
    for index, block_info in enumerate(self.block_tk_variables):
        def remove_block(instance=self, i=index):
            instance.block_tk_variables.pop(i)
            instance.refresh()

        row_frame = tk.Frame(self.block_frames)
        if row_frame.winfo_toplevel().tk.call('tk', 'windowingsystem') == 'aqua':
            padx = 0
            pady = 0
        else:
            padx = 2
            pady = 0
        row_frames.append(row_frame)
        row_frame.pack(expand=1, fill='both', padx=10, pady=2)

        opt_day, opt_hour, opt_duration = block_info

        btn_delete = tk.Button(row_frame, text="remove",  command=remove_block, padx=2*padx)
        om_day = ttk.Combobox(row_frame, textvariable=opt_day,  values=days, state="readonly" )
        om_time = entry_float(row_frame, textvariable=opt_hour)
        om_duration = entry_float(row_frame, textvariable=opt_duration)

        om_day.config(width=8)
        om_time.config(width=5)
        om_duration.config(width=5)

        om_day.pack(side='left', pady=pady, padx=padx)
        tk.Label(row_frame, text="start:", anchor='e',width=5,).pack(side='left', expand=1, fill='x')
        om_time.pack(side='left', pady=pady, padx=padx, expand=1, fill='x')
        tk.Label(row_frame, text="duration:", anchor='e',width=8,).pack(side='left', expand=1, fill='x')
        om_duration.pack(side='left', pady=pady, padx=padx, expand=1, fill='x')
        btn_delete.pack(side='left', pady=pady, padx=padx, ipadx=padx, expand=1, fill='x')

        om_time.bind("<Leave>", partial(_validate_start_time, opt_hour))
        om_time.bind("<FocusOut>", partial(_validate_start_time, opt_hour))
        om_duration.bind("<Leave>", partial(_validate_duration, opt_duration))
        om_duration.bind("<FocusOut>",  partial(_validate_duration, opt_duration))

# ================================================================================================================
# validate that the start time is a float, and that it is within operating hours
# ================================================================================================================
def _validate_start_time(tkvar: tk.StringVar,  _: tk.Event):
    """
    Validates the start time is good, and rounds to the nearest 15 minutes
    :param tkvar: The tk variable to analyze
    """
    try:
        value = float(tkvar.get())
    except ValueError:
        tk.messagebox.showerror(message="Class Start Time", title="Invalid Float",
                                detail=f"'{tkvar.get()}' is not a valid number")
        tkvar.set("8.0")
        return

    if value < 8 or value > 18:
        tk.messagebox.showerror(title="Invalid Class Start Time", detail="Resetting to 8:00 am",
                                message=f"'{tkvar.get()}' is outside of operational hours")
        tkvar.set("8.0")
        return

    tkvar.set(str(round(4*value)/4))

# ================================================================================================================
# validate that the duration is a float, and that it is within operating constraints
# ================================================================================================================
def _validate_duration(tkvar: tk.StringVar, _: tk.Event):
    """
    Validates the duration length is good, and rounds to the nearest 15 minutes
    :param tkvar: The tk variable to analyze
    """

    try:
        value = float(tkvar.get())
    except ValueError:
        tk.messagebox.showerror(message="Class Duration", title="Invalid Float",
                                detail=f"'{tkvar.get()}' is not a valid number")
        return

    if value < 0.5 or value > 8:
        tk.messagebox.showerror(title="Invalid Class Duration", detail="Resetting to 1.5 hrs",
                                message=f"'{tkvar.get()}' is not a valid class duration")
        return

    tkvar.set(str(round(4*value)/4))









from tkinter import *
from tkinter.ttk import Notebook

# ISSUE: when changing notebook tabs, the notebook page doesn't redraw
mw = Tk()
mw.title("Notebook Test")
mw.geometry(f"{800}x{800}")

main_page_frame = Frame(mw, borderwidth=1, relief='ridge')
main_page_frame.pack(side='top', expand=1, fill='both')

notebook = Notebook(main_page_frame)
notebook.pack(expand=1, fill='both')
frames = []

def tab_change(*_):
    index = notebook.index(notebook.select())
    print(index, frames[index])
    for widget in frames[index].winfo_children():
        print(widget)
        # widget.update()
        # widget.tkraise()
        # widget.event_generate("<FocusOut>")
        # widget.event_generate("<FocusIn>")
        # widget.event_generate("<MouseWheel>")
        # widget.event_generate("<Create>")
        widget.event_generate("<Expose>")
        # widget.event_generate("<Configure>")
        # widget.event_generate("<Key>")
        # widget.event_generate("<Motion>")
        # widget.event_generate("<Reparent>")
        # widget.event_generate("<Property>")
        # widget.event_generate("<Visibility>")
        # widget.event_generate("<Leave>")
        # widget.event_generate("<Enter>")
        # widget.event_generate("<Map>")
        # widget.event_generate("<KeyRelease>")
        # widget.event_generate("<Gravity>")
        # widget.event_generate("<Circulate>")

        # for widget1 in widget.winfo_children():
        #     widget1.update()
        #     widget1.focus()
    mw.update()
    mw.update_idletasks()


for index, info in enumerate(("one", "two", "three")):
    frame = Frame(mw)
    frames.append(frame)
    Label(frame, text=info).pack()
    notebook.add(frame, text=info)

notebook.bind("<<NotebookTabChanged>>", tab_change)

mw.mainloop()



import pytest

from src.scheduling_and_allocation.model import Teacher


# ============================================================================
# Stream
# ============================================================================
def test_dept_getter():
    teacher = Teacher("Jane", "Doe", "CompSci")
    assert teacher.department == "CompSci"


def test_dept_setter():
    teacher = Teacher("Jane", "Doe", "CompSci")
    teacher.department = "Mathematics"
    assert teacher.department == "Mathematics"


def test_firstname_getter():
    teacher = Teacher("Jane", "Doe", "CompSci")
    assert teacher.firstname == "Jane"


def test_lastname_getter():
    teacher = Teacher("Jane", "Doe", "CompSci")
    assert teacher.lastname == "Doe"


def test_release_setter_getter():
    teacher = Teacher("Jane", "Doe", "CompSci")
    teacher.release = 11.3
    assert teacher.release == 11.3


def test_string_representation_short():
    teacher = Teacher("Jane", "Doe", "CompSci")
    assert str(teacher) == f"{teacher.firstname} {teacher.lastname}"


def test_firstname_setter_good():
    """Verifies that firstname setter can set_default_fonts_and_colours a valid first name."""
    teach = Teacher("John", "Smith")
    new_f_name = "Bob"
    teach.firstname = new_f_name
    assert new_f_name == teach.firstname


def test_firstname_setter_with_spaces():
    """Verifies that firstname setter can set_default_fonts_and_colours a valid first name which has spaces."""
    teach = Teacher("John", "Smith")
    new_f_name = "Bob Blue"
    teach.firstname = new_f_name
    assert new_f_name == teach.firstname


def test_firstname_setter_bad():
    """Verifies that firstname setter doesn't change for invalid input (empty strings)."""
    teach = Teacher("John", "Smith")
    bad_name = " "
    teach.bad_name = bad_name
    assert teach.firstname == "John"


def test_lastname_setter_good():
    """Verifies that the lastname setter can set_default_fonts_and_colours a valid (non-empty) last name for the Teacher."""
    teach = Teacher("John", "Smith")
    new_l_name = "Forstinger"
    teach.lastname = new_l_name
    assert new_l_name == teach.lastname


def test_lastname_setter_with_spaces():
    """Verifies that the lastname setter can set_default_fonts_and_colours a valid
    (non-empty) last name (with spaces) for the Teacher."""
    teach = Teacher("John", "Smith")
    new_l_name = "Forstinger Blue"
    teach.lastname = new_l_name
    assert new_l_name == teach.lastname



from __future__ import annotations

import tkinter as tk
from dataclasses import dataclass, field
from functools import partial
from typing import Callable, Optional, ClassVar, Any
from enum import Enum

import re
from os import path
try:
    from ..modified_tk import FindImages, ToolBar, TkColours
except ImportError:
    from src.scheduling_and_allocation.modified_tk import FindImages, ToolBar, TkColours

# =====================================================================================================================
# Valid Menu type items
# =====================================================================================================================
class MenuType(Enum):
    """ensure that the MenuItem resource_type is valid"""
    Command = 'command'
    Separator = 'separator'
    Radiobutton = 'radiobutton'
    Checkbutton = 'checkbutton'
    Cascade = 'cascade'


# =====================================================================================================================
# A class to define a single menu item (menu items can have children though)
# =====================================================================================================================
@dataclass
class MenuItem:
    """Define a menu_item (resource_type, options, etc)"""

    # a dictionary of all menu_items that were created.  Key is the 'name' property of the menu item
    all_menu_items: ClassVar[dict[str, MenuItem]] = {}

    # -----------------------------------------------------------------------------------------------------------------
    # properties
    # -----------------------------------------------------------------------------------------------------------------
    label: str = ""                             # what is shown to the user
    tear_off: bool = 0                          # menu can be detached from window?
    menu_type: MenuType = MenuType.Command      # type of menu item
    name: str = ""                              # unique name to define your menu item
    accelerator: str = ""                       # just a string that defines short-cuts... JUST A STRING!
    command: Callable = lambda *_: None         # which function to call when menu item is clicked
    children: list[MenuItem] = field(default_factory=list)  # any kids?
    underline: bool = False                     # underline a character based on accelerator
    bool_variable: bool = True                  # if a checkbutton type, what is the default value

    # -----------------------------------------------------------------------------------------------------------------
    # can't remember what this is for... too scared to delete it
    # -----------------------------------------------------------------------------------------------------------------
    def __post_init__(self):
        MenuItem.all_menu_items[self.name] = self

    # -----------------------------------------------------------------------------------------------------------------
    # define which options are valid depending on the menu item type
    # -----------------------------------------------------------------------------------------------------------------
    def options_dict(self) -> dict[str, str]:
        """return a dictionary of options, appropriate to the menu resource_type"""
        options = dict()
        if self.menu_type != MenuType.Separator:
            options = {'label': self.label, 'accelerator': self.accelerator,
                       'command': self.command, 'underline': self.underline}
        return options

    # -----------------------------------------------------------------------------------------------------------------
    # add a child to this menu item
    # -----------------------------------------------------------------------------------------------------------------
    def add_child(self, menu_item: MenuItem):
        """add a new menu item as a sub-child of on existing menu item"""

        # if self is not a cascade resource_type menu item, then it should not have
        # children (i.e. non-cascade menu-items should not be parents)
        if self.menu_type != MenuType.Cascade:
            msg = f"Menu '{self.name}' is of resource_type '{self.menu_type}' and cannot support children"
            raise TypeError(msg)

        # procreate
        self.children.append(menu_item)


# =====================================================================================================================
# A tool bar item (names, commands, hints, etc)
# =====================================================================================================================
@dataclass
class ToolbarItem:
    """data defining a _toolbar item, command, hint, image, etc"""
    command: Optional[Callable] = lambda *_: None
    hint: Optional[str] = None
    image: Optional[str] = None


# =====================================================================================================================
# Generate a menu
# =====================================================================================================================
def generate_menu(mw, menu_details: Optional[list[MenuItem]], parent: tk.Menu):
    """
    Create a gui menu bar based off of the menu_details
    :param mw: main window to attach the menu to
    :param menu_details: list of menu items
    :param parent: the tk Menu (you have to have one of these first before you can attach menu items)
    :return:
    """

    if menu_details is None:
        return

    for menu_item in menu_details:

        # accelerator depends on window type
        if menu_item.accelerator:
            menu_item.accelerator = menu_item.accelerator.replace("Ctrl", "Control")
            if mw.tk.call('tk', 'windowingsystem') == 'aqua':
                if menu_item.accelerator:
                    menu_item.accelerator = menu_item.accelerator.replace("Control", "Command")

        # Cascading menu
        if menu_item.menu_type == MenuType.Cascade:
            new_menu = tk.Menu(parent, tearoff=menu_item.tear_off)
            generate_menu(mw, menu_item.children, new_menu)
            parent.add_cascade(label=menu_item.label, menu=new_menu)

        # simple separator
        elif menu_item.menu_type == MenuType.Separator:
            parent.add_separator()

        # button or command or radiobutton or...
        elif menu_item.menu_type == MenuType.Checkbutton:
            options: dict = menu_item.options_dict()
            options["variable"] = tk.BooleanVar(value= menu_item.bool_variable)
            options["command"] = partial(options["command"], options["variable"])
            parent.add_checkbutton(**options)

        else:
            options: dict = menu_item.options_dict()
            parent.add(menu_item.menu_type.value, **options)

        # bind accelerators
        if menu_item.accelerator:
            x = re.match(r'Control-(.)', menu_item.accelerator, re.RegexFlag.IGNORECASE)
            if x:
                binding = f"<Control-{x.group(1).lower()}>"
                mw.bind(binding, menu_item.command)

            x = re.match(r'Command-(.)', menu_item.accelerator, re.RegexFlag.IGNORECASE)
            if x:
                binding = f"<Command-{x.group(1).lower()}>"
                mw.bind(binding, menu_item.command)


# =====================================================================================================================
# Make a toolbar ( a bunch of buttons in a frame)
# =====================================================================================================================
def make_toolbar(mw, button_names: list[str], actions: dict[str:ToolbarItem],
                 colours: TkColours = TkColours(), **kwargs) -> ToolBar:
    """
    Create a toolbar described by the actions
    :param mw: the main window, or top level object, that the toolbar sits in
    :param button_names: name of the buttons
    :param actions: list of Toolbar items (name/tooltip/function)
    :param colours: default colours
    :param kwargs: any extra arguments that will passed to the toolbar widget
    :return:
    """
    image_dir = FindImages.get_image_dir()
    toolbar = ToolBar(mw, colours, **kwargs)

    # create all the buttons
    for button_name in button_names:
        if not button_name:  # if button not defined, insert a divider
            toolbar.bar()
        else:
            image = actions[button_name].image if actions[button_name].image else f'{button_name}.gif'
            full_image_path = path.join(image_dir, image)
            toolbar.add(
                name=button_name,
                image=full_image_path,
                command=actions[button_name].command,
                hint=actions[button_name].hint,
            )
    return toolbar



"""
# ============================================================================
# Create a Notebook
#
# EVENT HANDLERS
#
#   tab_changed_handler(tab_name, frame)
#
# ============================================================================

"""
from dataclasses import dataclass, field
from functools import partial
import tkinter.ttk as ttk
import tkinter as tk
from typing import Optional, Callable, Protocol, Any


@dataclass
class TabInfoProtocol(Protocol):
    name: str
    label: str
    subpages: list = field(default_factory=list)
    frame_args: dict[str, Any] = field(default_factory=dict)
    creation_handler: Callable = lambda *_: None
    is_default_page: bool = False


def _expose_widgets(widget):
    """weird shit happens on MAC in that widgets are not shown in tab... this forces them to be shown"""
    widget.event_generate("<Expose>")

    for child in widget.winfo_children():
        _expose_widgets(child)


class NoteBookFrameTk:
    """creates notebook, recursively, until all notebook tabs are created"""

    # ===================================================================================
    # constructor
    # ===================================================================================
    def __init__(self, mw: tk.Toplevel, main_page_frame: tk.Frame,
                 tabs_info: Optional[list[TabInfoProtocol]] = None,
                 tab_changed_handler: Callable[[str, tk.Frame], None] = lambda *_: None):
        """
        create the notebook(s)
        :param mw: toplevel window
        :param main_page_frame: where the current notebook frame should be placed
        :param tabs_info: what tabs do you need
        :param tab_changed_handler: who you gonna call?... ghost busters!
        """

        self._default_notebook_page = None
        self.tabs_info = tabs_info
        self.mw = mw
        self.frame = main_page_frame
        self.tab_frames: dict[tuple[ttk.Notebook, int], tuple[str, tk.Frame]] = {}
        self.tab_changed_handler = tab_changed_handler
        self.main_notebook_frame = None
        self.notebooks: list[ttk.Notebook] = []

        self.recursive_notebook_creation(self.frame, tabs_info)
        self.notebooks[0].select(self._default_notebook_page)
        self.mw.after(20, lambda: self.tab_changed(self.notebooks[0]))
        self.mw.after(10, lambda *_: self.mw.geometry(
            f"{max(975, self.mw.winfo_width())}x{max(700, self.mw.winfo_height())}"))


    # ===================================================================================
    # create notebook pages
    # ===================================================================================
    def recursive_notebook_creation(self, notebook_frame: tk.Frame, tabs_info: list[TabInfoProtocol]):
        """
        recursive calling of notebook creation
        :param notebook_frame:
        :param tabs_info:
        """

        if tabs_info is None:
            return

        # Create notebook if required
        if self.notebooks is None:
            self.notebooks: list[ttk.Notebook] = []
        notebook = ttk.Notebook(notebook_frame)
        self.notebooks.append(notebook)
        notebook.pack(expand=1, fill='both')

        if self.main_notebook_frame is None:
            self.main_notebook_frame = notebook

        for info in tabs_info:

            # create a frame and add to the notebook
            frame = tk.Frame(self.mw, **info.frame_args)
            notebook.add(frame, text=info.label)
            notebook_index = notebook.index(frame)

            # save the frame and name for this tab index
            self.tab_frames[(notebook, notebook_index)] = (info.name, frame)

            # set the page that should be shown by default
            if info.is_default_page:
                self._default_notebook_page = notebook_index

            # if this notebook page has sub-pages, then set them up as well
            if info.subpages:
                self.recursive_notebook_creation(frame, info.subpages)

            # call the 'create handler' for this page
            if info.creation_handler:
                info.creation_handler(info.name, frame)

        # add the binding for changing of events, and include a list of events for each tab change
        notebook.bind("<<NotebookTabChanged>>", partial(self.tab_changed, notebook))

    # ===================================================================================
    # tab has changed
    # ===================================================================================
    def tab_changed(self, notebook: ttk.Notebook, *_):
        """
        calls the tab changed callback when the tab has changed
        :param notebook:
        :param _: tk events?
        """
        try:
            index = notebook.index(notebook.select())
            tab_name, frame = self.tab_frames.get((notebook, index), None)
            self.tab_changed_handler(tab_name, frame)

            # have to delay to get this to work properly,
            # maybe using 'after_idle' works, but haven't tired it
            notebook.after(10, lambda *_:_expose_widgets(frame))
            notebook.after(20,lambda *_: notebook.event_generate("<Leave>"))

        except tk.TclError:
            """just in case tk has a problem with this as the notebook is being created"""
            pass




"""
Provides classes for managing scheduling conflicts
"""
from __future__ import annotations
import itertools
from typing import TYPE_CHECKING
from .enums import ConflictType

if TYPE_CHECKING:
    from .block import Block



# =============================================================================
# Constants
# =============================================================================
LUNCH_START: float = 11
LUNCH_END: float = 14
MAX_HOURS_PER_WEEK = 32.5


# -----------------------------------------------------------------------------------------------------------------
# set lunch break conflicts on a per day basis
# -----------------------------------------------------------------------------------------------------------------
def set_lunch_break_conflicts(blocks: tuple[Block, ...]):
    """
    Is there a lunch break for a teacher on each day
    :param blocks:
    :return: A list of conflicts
    """

   # collect blocks by day
    blocks_by_day = itertools.groupby(
        sorted(blocks, key=lambda b: b.day), lambda b: b.day)

    for _, blocks_group in blocks_by_day:

        daily_blocks: tuple[Block,...] = tuple(blocks_group)
        conflict = ConflictType.LUNCH if has_lunch_break_conflict(daily_blocks) else ConflictType.NONE

        # if no lunch on this day, create a conflict obj and mark blocks as conflicted_number
        if conflict == ConflictType.LUNCH:

            # find blocks that are the culprit
            for b in daily_blocks:
                if LUNCH_START <= b.start <= LUNCH_END:
                    _mark_block_conflict(ConflictType.LUNCH, (b,))
                if LUNCH_START <= b.end <= LUNCH_END:
                    _mark_block_conflict(ConflictType.LUNCH, (b,))
                if b.start <= LUNCH_START and b.end >= LUNCH_END:
                    _mark_block_conflict(ConflictType.LUNCH, (b,))


# -----------------------------------------------------------------------------------------------------------------
# are there any days within the given blocks where there is no lunch break
# -----------------------------------------------------------------------------------------------------------------
def has_lunch_break_conflict(blocks: tuple[Block, ...]) -> bool:
    """
    For this teacher, are there ANY days which do not have a lunch break?
    :param blocks:
    :return: yes or no (bool)
    """
    all_days_lunch = True

    # collect blocks by day
    blocks_by_day = itertools.groupby(
        sorted(blocks, key=lambda b: b.day), lambda b: b.day)

    for _, blocks_group in blocks_by_day:
        lunch_time = False
        blocks: list[Block] = list(blocks_group)
        blocks.sort(key=lambda b: b.start)

        # verify the 1st and last block
        b1: Block = blocks[0]
        b2: Block = blocks[-1]
        if b1.start - LUNCH_START > 0.49:
            continue
        if LUNCH_END - b2.end > 0.49:
            continue

        # look for a 1/2 window in-between blocks
        for b1, b2 in itertools.pairwise(blocks):

            # break between block
            start_break = max(b1.end, LUNCH_START)
            end_break = min(b2.start, LUNCH_END)
            if end_break > start_break and end_break - start_break > 0.49:
                lunch_time = True
                break

        all_days_lunch = all_days_lunch and lunch_time

    # if no lunch on this day, create a conflict obj and mark blocks as conflicted_number
    return not all_days_lunch


# -----------------------------------------------------------------------------------------------------------------
# are blocks overlapping in time/stream/lab/teacher
# -----------------------------------------------------------------------------------------------------------------
def set_block_conflicts(blocks_for_week: tuple[Block, ...],
                        conflict_type: ConflictType):
    """
    calculate if any two blocks overlap each other in time, and then
    assigns the conflict to the blocks
    :param blocks_for_week: a list of blocks to check for time overlap
    :param conflict_type: existing conflict type
    """
    blocks_by_day = itertools.groupby(
        sorted(blocks_for_week, key=lambda a: a.day), lambda a: a.day)

    for _, blocks in blocks_by_day:
        pairs: itertools.combinations[tuple[Block, Block]] = itertools.combinations(blocks, 2)
        for b1, b2 in pairs:
            if b1.conflicts_time(b2):
                _mark_block_conflict(ConflictType.TIME, (b1, b2))
                _mark_block_conflict(conflict_type, (b1, b2))


# -----------------------------------------------------------------------------------------------------------------
# are blocks overlapping in time/stream/lab/teacher
# -----------------------------------------------------------------------------------------------------------------
def block_conflicts_time(blocks_for_week: tuple[Block, ...]) -> ConflictType:
    """
    calculate if any two blocks overlap each other in time
    :param blocks_for_week: a list of blocks to check for time overlap
    :return: conflict_type
    """
    blocks_by_day = itertools.groupby(
        sorted(blocks_for_week, key=lambda a: a.day), lambda a: a.day)

    conflict = ConflictType.NONE
    for _, blocks in blocks_by_day:
        pairs: itertools.combinations[tuple[Block, Block]] = itertools.combinations(blocks, 2)
        for b1, b2 in pairs:
            if b1.conflicts_time(b2):
                conflict = conflict.TIME
    return conflict

# -----------------------------------------------------------------------------------------------------------------
# not enough days per week for a teacher
# -----------------------------------------------------------------------------------------------------------------
def set_number_of_days_conflict(blocks: tuple[Block, ...]):
    if has_number_of_days_conflict(blocks):
        _mark_block_conflict(ConflictType.MINIMUM_DAYS, blocks)

def has_number_of_days_conflict(blocks: tuple[Block, ...]) -> bool:
    """ collect blocks by day"""
    blocks_by_day = itertools.groupby(sorted(
        blocks, key=lambda a: a.day), lambda a: a.day)

    # if < 4 days, create a conflict and mark the blocks as conflicted_number
    if len(tuple(blocks_by_day)) < 4:
        return True
    return False


# -----------------------------------------------------------------------------------------------------------------
# too many hours for a week availability for a teacher
# -----------------------------------------------------------------------------------------------------------------
def set_availability_hours_conflict(blocks_for_teacher: tuple[Block, ...]):
    if has_availability_hours_conflict(blocks_for_teacher):
        _mark_block_conflict(ConflictType.AVAILABILITY, tuple(blocks_for_teacher))

def has_availability_hours_conflict(blocks: tuple[Block, ...]) -> bool:
    # collect blocks by day
    blocks_by_day = itertools.groupby(
        sorted(blocks, key=lambda a: a.day), lambda a: a.day)

    # if they have more than 32 hours worth of classes
    availability = 0
    for _, blocks_iter in blocks_by_day:
        blocks = list(blocks_iter)
        day_start = min(map(lambda b1: b1.start, blocks))
        day_end = max(map(lambda b1: b1.end, blocks))
        if day_end <= day_start:
            continue
        availability += day_end - day_start - 0.5

    # if over limit, create conflict
    if availability > MAX_HOURS_PER_WEEK:
        return True
    return False



def _mark_block_conflict(conflict_type: ConflictType, conflict_blocks: tuple[Block, ...]):
    for cb in conflict_blocks:
        cb.conflict = cb.conflict | conflict_type


# -----------------------------------------------------------------------------------------------------------------
# colors
# -----------------------------------------------------------------------------------------------------------------
def colours() -> dict[ConflictType, str]:
    """ Returns the colors used by each conflict resource_type """
    return ConflictType.colours()



import platform
import tkinter
import tkinter as tk
import tkinter.font

root = tk.Tk()
root.geometry("300x200")

e = tk.Entry(root)
my_default_font=e.cget("font")
default_font = tk.font.nametofont(my_default_font)
print(default_font.actual())


times = []
for h in range(0,24):
    for m in (0,15,30,45):
        times.append(f"{h:2d}:{m:02d}")
print(times)
time_index = 1

# Create multiple widgets
frame = tk.Frame(root, background='pink')
frame.pack()
the_time = tkinter.Variable()
entry = tk.Entry(frame, textvariable=the_time, width=10)
entry.pack(padx=10, pady=5,side='left')
word_text = tk.Text(frame, wrap='word', padx=10, pady=10, height=6, width=10)
word_text.pack( padx=10, pady=5, side='left')
word_text.config(state='disabled')

def _zoom(event: tk.Event=None):
    global time_index
    if event is not None:
        if event.delta > 0 or event.keysym=='Up' or event.keysym=="KP_Up":
            time_index -= 1
        elif event.delta < 0 or event.keysym=='Down'or event.keysym=="KP_Down":
            time_index+= 1
    time_index = time_index%len(times)
    the_time.set(times[time_index])

    # Insert text sections
    word_text.config(state='normal')
    word_text.delete('1.0','end')
    word_text.insert('end', times[(time_index-2)%len(times)]+'\n')
    word_text.insert('end', times[(time_index-1)%len(times)]+'\n')
    word_text.insert('end', times[time_index%len(times)]+'\n')
    word_text.insert('end', times[(time_index+1)%len(times)]+'\n')
    word_text.insert('end', times[(time_index+2)%len(times)])
    word_text.tag_add('small', '1.0', '1.end')
    word_text.tag_add('med', '2.0', '2.end')
    word_text.tag_add('big', '3.0', '3.end')
    word_text.tag_add('med', '4.0', '4.end')
    word_text.tag_add('small', '5.0', '5.end')
    word_text.tag_config('big', font='arial 20 bold', lmargin1=10)  # Set font, size and style
    word_text.tag_config('small', font='arial 10')  # Set font, size and style
    word_text.tag_config('med', lmargin1=5)  # Set font, size and style
    word_text.config(state='disabled')

    root.bell()
def what_key(event: tk.Event):
    print(event)

if platform.system() == "Linux":
    print ("linux")
    word_text.bind('<4>', _zoom)
    word_text.bind('<5>', _zoom)
else:
    word_text.bind('<MouseWheel>', _zoom)
    word_text.bind('<Key>', _zoom)

_zoom()

# Tag and style text sections


tk.mainloop()


"""Dialog box to edit a 'section' object"""
from __future__ import annotations

import tkinter as tk
import tkinter.messagebox as tk_message_box
import tkinter.simpledialog as simpledialog
from typing import Callable, Literal, TYPE_CHECKING

from ..gui_generics.number_validations import validate_int, entry_float, entry_int, validate_float
from .dialog_utilities import validate_class_times_equals_course_time, \
    get_block_info_from_tk_widgets, refresh_gui_blocks

from ..gui_generics.add_remove_tk import AddRemoveTk
if TYPE_CHECKING:
    from ..model import Lab, Teacher

# =====================================================================================================================
# Edit Course
# - add sections with blocks, teachers, labs
# =====================================================================================================================
class EditCourseDialogTk(simpledialog.Dialog):
    def __init__(self, frame:tk.Frame,
                 edit_or_add: Literal['edit','add'],
                 existing_course_numbers: list = None,
                 *,
                 course_number:str = "",
                 course_name: str = "",
                 course_hours: float = 3.0,
                 course_allocation: bool = True,
                 num_sections: int = 1,
                 assigned_teachers: list[Teacher] = None,
                 non_assigned_teachers: list[Teacher],
                 assigned_labs: list[Lab] = None,
                 non_assigned_labs: list[Lab],
                 current_blocks: list[tuple[str,float,float]] = None,
                 apply_changes: Callable[[str, str, float, bool, int, list, list, list], None]):
        """
        Edit Course  - add sections with blocks, teachers, labs
        :param frame: the parent of the dialog box
        :param edit_or_add: editing or adding a course?
        :param existing_course_numbers: all of the existing course numbers (to prevent duplication)
        :param course_number: course number
        :param course_name: course name
        :param course_hours: hours per week
        :param course_allocation: does the course require a teacher?
        :param num_sections: number of sections of this course
        :param assigned_teachers: which teachers have been assigned to this course
        :param non_assigned_teachers: which teachers have NOT been assigne to this course
        :param assigned_labs: which labs have been assigned to this course
        :param non_assigned_labs: which labs have not been assigned to this course
        :param current_blocks: What blocks (class times) have been assigned
        :param apply_changes: what to call when the user is finished updating their request
        """

        self.top_frame = frame
        self.edit_or_add = edit_or_add
        self.existing_course_numbers = [] if existing_course_numbers is None else existing_course_numbers
        self.course_number_tk = tk.StringVar(value = course_number)
        self.course_name_tk = tk.StringVar(value = course_name)
        self.course_hours_tk = tk.StringVar(value=str(course_hours))
        self.course_allocation_tk = tk.BooleanVar(value=course_allocation)
        self.num_sections_tk = tk.StringVar(value = str(num_sections))
        self._assigned_teachers = assigned_teachers if assigned_teachers is not None else []
        self._non_assigned_teachers = non_assigned_teachers
        self._assigned_labs = assigned_labs  if assigned_labs is not None else []
        self._non_assigned_labs = non_assigned_labs
        self.current_blocks = current_blocks if current_blocks is not None else []
        self._apply_changes = apply_changes


        self.block_tk_variables = []
        self.block_frames = None
        self.course_name = course_name

        dialog_title = "Edit Course" if self.edit_or_add == "edit" else "Add Course"
        super().__init__(frame.winfo_toplevel(), dialog_title)


    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame:tk.Frame):

        # ------------------------------------------------------------------------------------------------------------
        # Course
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame = tk.Frame(frame)
        tk.Label(course_info_frame, text="Course number", anchor='e', width=15).grid(row=0, column=0, sticky='nsew')
        en_course_number = tk.Entry(course_info_frame, textvariable=self.course_number_tk,)
        en_course_number.grid(row=0, column=1, sticky='nsew')
        if self.edit_or_add == 'edit':
            en_course_number.config(state='readonly')
            en_course_number.config(readonlybackground="#bbbbbb", foreground="#555555")

        tk.Label(course_info_frame, text="Course name", anchor='e', width=15).grid(row=1, column=0, sticky='nsew')
        en_course_name = tk.Entry(course_info_frame, textvariable=self.course_name_tk,)
        en_course_name.grid(row=1, column=1, sticky='nsew')

        tk.Label(course_info_frame, text="Hours per week", anchor='e', width=15).grid(row=2, column=0, sticky='nsew')
        en_course_name = entry_float(course_info_frame, textvariable=self.course_hours_tk)
        en_course_name.grid(row=2, column=1, sticky='nsew')

        tk.Label(course_info_frame, text="Needs Allocation", anchor='e', width=15).grid(row=3, column=0, sticky='nsew')
        en_allocation = tk.Checkbutton(course_info_frame, text="", offvalue=False, onvalue=True,
                                       variable=self.course_allocation_tk, width=20, justify='left')
        en_allocation.grid(row=3, column=1, sticky='w')

        # ------------------------------------------------------------------------------------------------------------
        # Sections/Blocks
        # ------------------------------------------------------------------------------------------------------------
        tk.Label(course_info_frame, text="Number of Sections", anchor='e', width=20).grid(row=4, column=0, sticky='nsew')
        description = entry_int(course_info_frame, textvariable=self.num_sections_tk)
        description.grid(row=4, column=1, sticky='nsew')

        self.block_frames = tk.Frame(frame)
        for index, block_info in enumerate(self.current_blocks):
            opt_day = tk.StringVar(value=block_info[0])
            opt_hour = tk.StringVar(value=str(block_info[1]))
            opt_duration = tk.StringVar(value=str(block_info[2]))
            self.block_tk_variables.append((opt_day, opt_hour, opt_duration))

        # ------------------------------------------------------------------------------------------------------------
        # Teacher/Lab/Stream Add/Remove
        # ------------------------------------------------------------------------------------------------------------
        teacher_assignments_frame = tk.Frame(frame)
        AddRemoveTk(teacher_assignments_frame, self._non_assigned_teachers, self._assigned_teachers,
                    "Assign Teacher to all Classes", "Remove Teacher from all Classes",height=10)

        lab_assignments_frame = tk.Frame(frame)
        AddRemoveTk(lab_assignments_frame, self._non_assigned_labs, self._assigned_labs,
                     "Assign Lab to all Classes","Remove Lab from all Classes", height=10)

        # ------------------------------------------------------------------------------------------------------------
        # layout
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame.grid(row=0, column=0)
        self.block_frames.grid(row=1, column=0, sticky='nsew', padx=5, pady=5)
        teacher_assignments_frame.grid(row=2,column=0, sticky='nsew', padx=5, pady=5)
        lab_assignments_frame.grid(row=3,column=0, sticky='nsew', padx=5, pady=5)

        self.refresh()

        return description

    # ================================================================================================================
    # add a new block
    # ================================================================================================================
    def add_new_block(self):
        self.block_tk_variables.append((tk.StringVar(value="Monday"), tk.StringVar(value="8.0"), tk.StringVar(value="1.5")))
        self.refresh()

    # ================================================================================================================
    # Update gui to reflect current info
    # ================================================================================================================
    def refresh(self):
        refresh_gui_blocks(self)
        tk.Button(self.block_frames, text="Add New Class Time", command=self.add_new_block, padx=5).pack(expand=1, fill='y')

    # ================================================================================================================
    # validate before applying
    # ================================================================================================================
    def validate(self):
        """
        Is the data, as entered by the user, valid?
        :return: True if data is good (changes are applied), false otherwise (nothing happens)
        """
        if self.edit_or_add == 'add':
            if self.course_number_tk.get() in self.existing_course_numbers:
                tk_message_box.showerror("Course Number",f"Course number '{self.course_number_tk.get()}' already exists")
                return False
        if not validate_float(self.course_hours_tk.get(),"Course hours", "The number of hours per week must be a valid float!"):
            return False

        if not validate_int(self.num_sections_tk.get(),"Number of Sections", "The number of sections must be an integer!"):
            return False

        if not validate_class_times_equals_course_time(self.block_tk_variables, float(self.course_hours_tk.get())):
            return False

        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        """apply the changes and close the dialog"""
        new_blocks=get_block_info_from_tk_widgets(self.block_tk_variables)
        self._apply_changes(self.course_number_tk.get(), self.course_name_tk.get(),
                            float(self.course_hours_tk.get()),self.course_allocation_tk.get(),
                            int(self.num_sections_tk.get()), self._assigned_teachers,
                    self._assigned_labs, new_blocks)




from tkinter import *

MOVABLE_TAG_NAME = "movable"


def main():
    mw: Tk = Tk()

    canvas: Canvas = Canvas(mw)
    canvas.pack(expand=1, fill='both')


    canvas.tag_bind(MOVABLE_TAG_NAME, "<Button-1>", lambda e: select_gui_block_to_move(canvas, e))

    r1: int = canvas.create_rectangle(50, 50, 150, 150, fill='blue', tags=(MOVABLE_TAG_NAME, "obj1"))
    s1: int = canvas.create_text(100, 100, text="A", tags=(MOVABLE_TAG_NAME, "obj1"))
    r2: int = canvas.create_rectangle(70, 70, 170, 170, fill='red', tags=(MOVABLE_TAG_NAME, "obj2"))
    s2: int = canvas.create_text(120, 120, text="B", tags=(MOVABLE_TAG_NAME, "obj2"))
    mw.mainloop()


def select_gui_block_to_move(canvas: Canvas, event: Event):
    obj: tuple[int, ...] = canvas.find_withtag('current')
    print(obj)
    current_tags = canvas.gettags(obj[0])
    print(current_tags)
    tags: tuple[str] = tuple(filter(lambda x: x != MOVABLE_TAG_NAME and x != 'current', canvas.gettags(obj[0])))
    print(tags)
    canvas.tag_raise(tags, 'all')

    # unbind any previous binding for clicking and motion, just in case
    canvas.bind("<Motion>", "")
    canvas.bind("<ButtonRelease-1>", "")
    canvas.tag_bind(MOVABLE_TAG_NAME, "<Button-1>", "")

    # bind for mouse motion
    canvas.bind("<Motion>",
                lambda e: gui_block_is_moving(canvas, tags[0], event.x, event.y, e))

    # bind for release of mouse up
    canvas.bind("<ButtonRelease-1>", lambda e: gui_block_has_stopped_moving(canvas, tags[0], e))


def gui_block_is_moving(canvas, tag, x, y, event):
    # unbind moving while we process
    canvas.bind("<Motion>", "")

    # move the widget
    canvas.move(tag, event.x - x, event.y - y)

    # rebind for motion
    canvas.bind("<Motion>",
                lambda e: gui_block_is_moving(canvas, tag, event.x, event.y, e))


def gui_block_has_stopped_moving(canvas, tag, event):
    canvas.tag_bind(MOVABLE_TAG_NAME, "<Button-1>", lambda e: select_gui_block_to_move(canvas, e))
    canvas.bind("<Motion>", "")
    canvas.bind("<ButtonRelease-1>", "")


if __name__ == "__main__":
    main()




"""
# ============================================================================
# This module contains all the gui code that is related to the course tree view
# - There is a course tree
# - There is three lists of resources (for drag'n'dropping onto the tree)
#
# EVENT HANDLERS
#
#   double-click: edit the selected tree object
#       handler_tree_edit(selected_obj, parent_obj, tree_id, parent_id)
#
#   button: create a new course (button)
#       handler_new_course()
#
#   tree right-click: return drop-down menu info
#       handler_tree_create_popup(selected_obj, parent_obj, tree_id, parent_id)->list[MenuItem]
#
#   resource right-click: return drop-down menu info
#       handler_resource_create_menu(resource_type, object)->list[MenuItem]
#
#   teacher double-click:
#       handler_show_teacher_stat(teacher)
#
#   resource object being dragged over tree, return bool to indicate valid drop site
#       handler_drag_resource(resource_type, tree_object) -> bool
#
#   resource object dropped on tree object
#       handler_drop_resource(resource_obj, tree_obj, tree_id)
#
# ============================================================================
"""
from __future__ import annotations

import time
import tkinter as tk
from functools import partial
from tkinter.messagebox import showinfo
from typing import Callable, Any, TYPE_CHECKING
import re

from ..modified_tk import Scrolled
from ..modified_tk import AdvancedTreeview
from ..modified_tk import DragNDropManager
from ..modified_tk.InitGuiFontsAndColours import get_fonts_and_colours, set_default_fonts
from ..gui_generics.menu_and_toolbars import MenuItem, MenuType, generate_menu
from ..model import ResourceType

if TYPE_CHECKING:
    from ..model import Teacher, Lab, Stream

    RESOURCE_OBJECT = [Teacher | Lab | Stream]
    TREE_OBJECT = Any


# ====================================================================================================================
# default stuff
# ====================================================================================================================
def _default_menu(obj, parent_obj, iid, parent_iid) -> list[MenuItem]:
    menu = MenuItem(name=str(obj), label=str(obj), menu_type=MenuType.Command, command=lambda: None)
    return [menu, ]


def _default_resource_menu(resource_type: ResourceType, obj) -> list[MenuItem]:
    menu_title = MenuItem(name=str(resource_type), label=str(resource_type), menu_type=MenuType.Command,
                          command=lambda: None)
    menu = MenuItem(name=str(obj), label=str(obj), menu_type=MenuType.Command, command=lambda: None)
    return [menu_title, menu]


# ====================================================================================================================
# Edit courses GUI_Pages
# ====================================================================================================================

class EditCoursesTk:
    """A page that allows user to edit courses, assign resources, sections, blocks, etc
    :param frame: the frame to hold all the gui elements

    """

    # -----------------------------------------------------------------------------------------------------------------
    # constructor
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, frame: tk.Frame):
        """create the EditCourse page
        :param frame: container object for the gui stuff
        """
        self.frame = frame

        # setup fonts if they have not already been set up
        self.colours, self.fonts = get_fonts_and_colours()
        if self.fonts is None:
            self.colours, self.fonts = set_default_fonts(self.frame.winfo_toplevel())

        # call backs (should be defined by presenter)
        self.handler_tree_edit: Callable[[TREE_OBJECT, TREE_OBJECT, str, str], None] \
            = lambda obj, parent_obj, tree_id, parent_id: print(f"Edit {str(obj)}")
        self.handler_new_course: Callable[[], None] = lambda: None
        self.handler_tree_create_popup: Callable[[TREE_OBJECT, TREE_OBJECT, str, str], list[MenuItem]] = _default_menu
        self.handler_resource_create_menu: Callable[
            [ResourceType, RESOURCE_OBJECT], list[MenuItem]] = _default_resource_menu
        self.handler_show_teacher_stat: Callable[[RESOURCE_OBJECT], None] = lambda obj: print(f"Teacher stat: {obj}")
        self.handler_drag_resource: Callable[[ResourceType, TREE_OBJECT], bool] = \
            lambda resource_type, target_obj: True
        self.handler_drop_resource: Callable[[ RESOURCE_OBJECT, TREE_OBJECT, str], None] = \
            lambda source_obj, target_obj, tree_id: None

        # create lists to keep track of widgets and objects for all resources
        # (teachers/streams/labs)
        self.resource_Listbox: dict[ResourceType, tk.Listbox] = dict()
        self.resource_objects: dict[ResourceType, list[RESOURCE_OBJECT]] = dict()
        for rt in ResourceType:
            self.resource_objects[rt] = list()

        # remove anything that is already in the frame
        for widget in frame.winfo_children():
            widget.destroy()

        # make panels
        right_panel = tk.Frame(frame)
        right_panel.grid(row=0, column=1, sticky='nsew')
        left_panel = tk.Frame(frame)
        left_panel.grid(row=0, column=0, sticky='nsew')

        # calculate min_width of left panel based on screen size
        width = 975
        height = 800
        if match:= re.match(r"^=?(\d+)x(\d+)?([+-]\d+[+-]\d+)?$", frame.winfo_toplevel().geometry()):
            width, height, _ = match.groups()

        # relative weights etc. to widths
        frame.grid_columnconfigure(0, minsize=int(0.5 * int(width)), weight=1)
        frame.grid_columnconfigure(1, weight=1)
        frame.grid_rowconfigure(0, weight=1)

        # make the gui contents
        self.course_ttkTreeView = self._make_treeview(left_panel)
        self._make_resource_list_widgets(right_panel)
        self._make_button_row(right_panel)

        right_panel.grid_columnconfigure(0, weight=1)
        right_panel.grid_rowconfigure(0, weight=10, minsize=int(0.45 * int(height)))
        right_panel.grid_rowconfigure(1, weight=1)
        right_panel.grid_rowconfigure(2, weight=1)
        right_panel.grid_rowconfigure(3, weight=0)


        # is the user hovering over a tree element when dragging a resource?
        self._hover_time = None
        self._hover_tree_element = None


    # -----------------------------------------------------------------------------------------------------------------
    # Public Methods for manipulating the tree object
    # -----------------------------------------------------------------------------------------------------------------
    def update_resource_type_objects(self, resource_type: ResourceType, objs: tuple[RESOURCE_OBJECT, ...]):
        """
        Updates the Listbox widget with Labs/Teachers/Streams string representations
        :param resource_type: what view resource_type are you updating?
        :param objs: all the objects that you want in the list
        """
        widget: tk.Listbox = self.resource_Listbox[resource_type]
        widget.delete(0, 'end')
        self.resource_objects[resource_type].clear()
        for obj in objs:
            widget.insert('end', str(obj))
            self.resource_objects[resource_type].append(obj)

    def add_tree_item(self, parent_id: str, name: str, child: TREE_OBJECT, hide: bool = True) -> str:
        """add an object to the tree, as a child of the parent
        :param parent_id: an existing tree object that will become the parent of this new child
        :param name: the name of
        :param child: the child
        :param hide: hide the children?
        :return: The id of the tree element
        """
        tag = "bold" if parent_id is None or parent_id == "" else "normal"
        parent_id = "" if parent_id is None else parent_id
        iid = self.course_ttkTreeView.insert_sorted(parent_id, child, text=name, tag=tag)
        self.course_ttkTreeView.after(20,lambda *_: self.course_ttkTreeView.item(parent_id, open=not hide))
        return iid

    def remove_tree_item(self, tree_iid: str):
        """remove an item from the tree
        :param tree_iid: the id of the item to be removed"""
        return self.course_ttkTreeView.delete(tree_iid)

    def remove_tree_item_children(self, tree_iid: str):
        """remove all children from the specified parent
        :param tree_iid: the id of the parent whose items are to be removed"""
        for child in self.course_ttkTreeView.get_children(tree_iid):
            self.course_ttkTreeView.delete(child)

    def update_tree_text(self, tree_iid: str, text: str):
        if self.course_ttkTreeView.exists(tree_iid):
            self.course_ttkTreeView.item(tree_iid, text=text)

    def clear_tree(self):
        """clear the tree of all items"""
        for child in self.course_ttkTreeView.get_children():
            self.course_ttkTreeView.delete(child)

    # -----------------------------------------------------------------------------------------------------------------
    # Resource Event - pop-up menu for teacher/stream/lab
    # -----------------------------------------------------------------------------------------------------------------
    def _cmd_show_resource_type_menu(self, resource_type: ResourceType, e: tk.Event):
        """bound method for right click on resource list
                calls handler: handler_resource_create_menu
        """

        if resource_type is ResourceType.none:
            return
        widget: tk.Listbox = self.resource_Listbox[resource_type]
        index = widget.nearest(e.y)
        widget.select_clear(0, 'end')
        widget.selection_set(index, index)
        obj = self.resource_objects[resource_type][index]

        # get menu info from callback routine
        menu = tk.Menu(self.frame.winfo_toplevel(), tearoff=0)
        menu_details = self.handler_resource_create_menu(resource_type, obj)
        generate_menu(self.frame.winfo_toplevel(), menu_details, menu)
        try:
            menu.tk_popup(e.x_root, e.y_root)
        finally:
            menu.grab_release()

    # -----------------------------------------------------------------------------------------------------------------
    # Resource Event - double click teacher on resource view
    # -----------------------------------------------------------------------------------------------------------------
    def _cmd_double_click_teacher(self, e: tk.Event):
        """bound method for double-click on a teacher object,
                calls handler: handler_show_teacher_stat
        """
        widget: tk.Listbox = self.resource_Listbox[ResourceType.teacher]
        index = widget.nearest(e.y)
        widget.select_clear(0, 'end')
        widget.selection_set(index, index)
        obj = self.resource_objects[ResourceType.teacher][index]
        self.handler_show_teacher_stat(obj)

    # -----------------------------------------------------------------------------------------------------------------
    # Tree Event - show editor for selected element (course, stream, block)
    # -----------------------------------------------------------------------------------------------------------------
    def _cmd_edit_selection(self, *_):
        """bound method for <keypress Enter> or <double click> on treeview, or clicking 'Edit Selection' button
                calls handler: handler_tree_edit
        """

        # get the tree selected item
        iid_list = self.course_ttkTreeView.selection()
        if not iid_list:
            return
        obj = self.course_ttkTreeView.get_obj_from_id(iid_list[0])
        iid = iid_list[0]
        parent_iid = self.course_ttkTreeView.parent(iid)

        # get object associated with this item
        obj = self.course_ttkTreeView.get_obj_from_id(iid)
        parent_obj = self.course_ttkTreeView.get_obj_from_id(parent_iid)

        self.handler_tree_edit(obj,parent_obj,  iid, parent_iid)

    # -----------------------------------------------------------------------------------------------------------------
    # Tree Events - pop-up menu for course/stream/block
    # -----------------------------------------------------------------------------------------------------------------
    def _cmd_show_tree_menu(self, e: tk.Event):
        """bound method for right click on tree item
                calls handler: handler_tree_create_popup
        """

        tv = self.course_ttkTreeView

        # which item was selected?
        iid = tv.identify_row(e.y)
        if not iid:
            return
        tv.selection_set(iid)
        parent_iid = tv.parent(iid)

        # get object associated with this item
        obj = tv.get_obj_from_id(iid)
        parent_obj = tv.get_obj_from_id(parent_iid)

        # get menu info from callback routine
        menu = tk.Menu(self.frame.winfo_toplevel(), tearoff=0)
        menu_details = self.handler_tree_create_popup(obj, parent_obj, iid, parent_iid)
        generate_menu(self.frame.winfo_toplevel(), menu_details, menu)
        try:
            menu.tk_popup(e.x_root, e.y_root)
        finally:
            menu.grab_release()

    # -----------------------------------------------------------------------------------------------------------------
    # Event Drag 'n' Drop start
    # -----------------------------------------------------------------------------------------------------------------
    def _resource_on_start_drag(self, e: tk.Event, info_data: dict) -> str:
        """bound event if a resource widget is about to be dragged from a resource listbox"""

        # what resource type is it?
        info_data["resource_type"] = ResourceType.none
        lb: tk.Listbox = e.widget
        rtype: ResourceType = ResourceType.none
        for rt in self.resource_Listbox:
            if lb == self.resource_Listbox[rt]:
                rtype = rt

        # select the nearest item in the list
        index = lb.nearest(e.y)
        lb.select_clear(0, 'end')
        lb.selection_set(index, index)

        # save some info for later use
        info_data["source_obj"] = self.resource_objects[rtype][index]
        info_data["resource_type"] = rtype

        # return string to be used in the drag indicator
        return lb.get(index)

    # -----------------------------------------------------------------------------------------------------------------
    # Event Drag 'n' Drop moving
    # -----------------------------------------------------------------------------------------------------------------
    def _hovering_response(self, iid):
        tv = self.course_ttkTreeView
        tv.item(iid, open=True)

    def _resource_on_drag(self, e: tk.Event, info_data: dict, target: tk.Widget) -> bool:
        """bound event while the resource list object is being dragged around
                calls handler: handler_drag_resource"""

        tv = self.course_ttkTreeView
        for s_iid in tv.selection():
            tv.selection_remove(s_iid)

        tv_y_pos = e.y_root - tv.winfo_rooty()

        # a resource widget can only be dropped on the self.course_ttkTreeView
        if target == tv:

            # get position of mouse within the treeview widget, and identify item in treeview under mouse
            tv_y_pos = e.y_root - tv.winfo_rooty()
            iid = tv.identify_row(tv_y_pos)

            # if we have a valid id...
            if iid:
                target_obj: TREE_OBJECT = tv.get_obj_from_id(iid)
                resource_type: ResourceType = info_data["resource_type"]

                # check with 'user' if this is a valid item for dropping source onto target
                if self.handler_drag_resource(resource_type, target_obj):

                    # if we are hovering, then open the tree element
                    if (self._hover_time is not None and self._hover_tree_element is not None
                            and self._hover_tree_element != iid):
                        self.frame.after_cancel(self._hover_time)

                    if self._hover_tree_element is None or self._hover_tree_element != iid:
                        self._hover_tree_element = iid
                        self._hover_time = self.frame.after(500, self._hovering_response, iid)

                    info_data["target_obj"] = target_obj
                    info_data["tree_id"] = iid
                    tv.selection_set(iid)
                    return True

        # if mouse above or below tv view, then force scroll, but have a time delay to make it more
        # manageable
        elif tv_y_pos < 0:
            if info_data.get('start_time', None) is not None and time.time() - info_data['start_time'] < 0.05:
                return False
            info_data['start_time'] = time.time()
            n1,n2 = self.tree_scrolled.yview_widget()
            self.tree_scrolled.yview_moveto(max(7/8*n1 - 1/8*n2,0))
        elif tv_y_pos > 0:
            if info_data.get('start_time', None) is not None and time.time() - info_data['start_time'] < 0.05:
                return False
            info_data['start_time'] = time.time()
            n1,n2 = self.tree_scrolled.yview_widget()
            self.tree_scrolled.yview_moveto(min(7/8*n1 + 1/8*n2,1))

        return False

    # -----------------------------------------------------------------------------------------------------------------
    # Event Drag 'n' Drop end
    # -----------------------------------------------------------------------------------------------------------------
    def _resource_on_drop(self, e: tk.Event, info_data: dict, target: tk.Widget):
        """bound event when a resource item is dropped
                calls handler: handler_drop_resource"""

        if self._resource_on_drag(e, info_data, target):
            target_obj: TREE_OBJECT = info_data["target_obj"]
            source_obj: RESOURCE_OBJECT = info_data["source_obj"]
            tree_id: str = info_data["tree_id"]
            self.handler_drop_resource(source_obj, target_obj, tree_id )
        if self._hover_time is not None:
            self.frame.after_cancel(self._hover_time)
        self._hover_time = None
        self._hover_tree_element = None

    # -----------------------------------------------------------------------------------------------------------------
    # Button Events - new course
    # -----------------------------------------------------------------------------------------------------------------
    def _new_course_btn_pressed(self, *_):
        """bound method for clicking 'new course' button
                calls handler: handler_new_course"""

        self.handler_new_course()

    # =================================================================================================================
    # Private Methods
    # =================================================================================================================

    def _make_treeview(self, left_panel: tk.Frame):
        """
        make tree representing all courses, sections, blocks, resource
        :param left_panel: where to put stuff
        """
        self.tree_scrolled: Scrolled = Scrolled(left_panel, 'AdvancedTreeview', scrollbars='se')
        self.tree_scrolled.pack(expand=1, fill='both', side='left')
        tv: AdvancedTreeview = self.tree_scrolled.widget
        tv.bind('<Double-1>', self._cmd_edit_selection)
        tv.bind('<Key-Return>', self._cmd_edit_selection)
        tv.tag_configure("bold", font=self.fonts.big)
        tv.tag_configure("normal", font=self.fonts.normal)

        # mac vs windows methods for getting pop up menus
        if self.frame.winfo_toplevel().tk.call('tk','windowingsystem') == 'aqua':
            tv.bind('<2>', self._cmd_show_tree_menu)
            tv.bind('<Control-1>', self._cmd_show_tree_menu)
        else:
            tv.bind('<3>', self._cmd_show_tree_menu)

        return tv

    def _make_resource_list_widgets(self, panel):
        """make a list widget for teachers/labs/streams"""

        # the scrolled method requires an empty frame, or else, it messes up.
        f = tk.Frame(panel)
        f.grid(column=0, stick='nsew', row=0)
        tk.Label(f, text="Teachers", font=self.fonts.bold).pack()
        sf = tk.Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='e')
        s.widget.bind('<Double-Button-1>', self._cmd_double_click_teacher)
        self.resource_Listbox[ResourceType.teacher] = s.widget

        f = tk.Frame(panel)
        f.grid(column=0, stick='nsew', row=1)
        tk.Label(f, text="Labs", font=self.fonts.bold).pack()
        sf = tk.Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='e')
        self.resource_Listbox[ResourceType.lab] = s.widget

        f = tk.Frame(panel)
        f.grid(column=0, stick='nsew', row=2)
        tk.Label(f, text="Streams", font=self.fonts.bold).pack()
        sf = tk.Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='e')
        self.resource_Listbox[ResourceType.stream] = s.widget

        # Set up binding for menus and dragging-n-dropping
        dm = DragNDropManager()

        for resource_type in ResourceType:
            if resource_type not in self.resource_Listbox:
                continue
            dm.add_draggable(self.resource_Listbox[resource_type],
                             on_start=self._resource_on_start_drag,
                             on_drop=self._resource_on_drop,
                             on_drag=self._resource_on_drag)

            # mac vs windows methods for getting pop up menus
            if self.frame.winfo_toplevel().tk.call('tk', 'windowingsystem') == 'aqua':
                self.resource_Listbox[resource_type].bind(
                    '<Button-2>',
                    partial(self._cmd_show_resource_type_menu, resource_type))
                self.resource_Listbox[resource_type].bind(
                    '<Control-1>',
                    partial(self._cmd_show_resource_type_menu, resource_type))
            else:
                self.resource_Listbox[resource_type].bind(
                    '<Button-3>',
                    partial(self._cmd_show_resource_type_menu, resource_type))

            self.resource_Listbox[resource_type].unbind('<Motion>')

    def _make_button_row(self, panel: tk.Frame):
        """make the button row for creating new courses, editing selections, etc"""
        button_row = tk.Frame(panel)
        button_row.grid(column=0, sticky='nsew', row=3)
        btn_new_course = tk.Button(button_row, text="New Course", width=11,
                                command=lambda: self._new_course_btn_pressed())
        btn_new_course.pack(side='left', ipadx=10)
        btn_edit_selection = tk.Button(button_row, text="Edit Selection", width=11,
                                    command=lambda: self._cmd_edit_selection())
        btn_edit_selection.pack(side='left', ipadx=10)

    # ================================================================================================================
    # show a message
    # ================================================================================================================
    def show_message(self, title: str, msg: str, detail: str = ""):
        showinfo(title=title, message=msg, detail=detail, icon='info')
